import matplotlib
from .lipid import Lipid, FA
from .reference import ReferenceLipid, parse_lipid_reference_table, check_reference, parse_lipid_reference_table_dict
from .validation import reaction_validator
from .reaction import Reaction, FAReaction
from .edgelist import Edge, Edgelist, reaction_str, enzyme_str
from .reaction_evaluation import Extender
from .GenerateLipidNetwork import GenerateLipidNetwork
from .parser import parse_fa_reactions, parse_excluded_fa_reactions, fa_parser, \
    parse_fatty_acids, save_lipidlynxx_converter, lipid_parser, parse_lipid_list

from .LipidNetwork import LipidNetwork
from .vis_utils import STATIC_NODE_PROPERTIES
from ._network_lipidome_summary import plot_pca, plot_heatmap


# utilities to show available attributes for plotting
def node_colour_options() -> set:
    return LipidNetwork._allowed_node_attributes_ - STATIC_NODE_PROPERTIES


def edge_colour_options() -> set:
    return LipidNetwork._allowed_edge_attributes_


def node_size_options() -> set:
    return {
        attr for attr in
        LipidNetwork._allowed_node_attributes_ - STATIC_NODE_PROPERTIES
        if not LipidNetwork._discrete_map_.get(attr, True)
    }


# Temporary functions for parsing curated Reactome and Rhea tables
from .tmp_dirty_functions import make_organism_reaction_list_from_reactome, make_all_reaction_list_from_reactome, \
    extender, make_all_reaction_list_from_rhea




# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
import os
from .misc import _tuple_to_string_
from typing import Union, Dict, Any, Tuple, List
from jinja2 import Template
from pyvis.network import Network
import networkx as nx
import webbrowser
import numpy as np
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize, to_hex
from warnings import warn
import json
from itertools import combinations

# NOTE: if you're adding a new attribute here, make sure
#       you also add it to LipidNetwork._allowed_<node/edge>_attributes_
#       otherwise it will not be available. If the newly added feature
#       is not static, it will be necessary to manually add it to the
#       LipidNetwork instance by calling add_attributes_to_plot
# Auxiliaries for computing annotations
STATIC_EDGE_PROPERTIES = {
    # edge attributes
    "reaction_type",
    # "enzyme_id",
    # "enzyme_gene_name",
    # "enzyme_uniprot"
}
STATIC_NODE_PROPERTIES = {
    # node attributes
    "lipid_class",
    "chain_length",
    "desaturation",
    "hydroxylation",
    "c_index",
    "db_index",
    "oh_index",
    "degree",
    "betweenness",
    "closeness"
}
STATIC_NON_LIPID_PROPERTIES = {
    # node attributes
    "degree",
    "betweenness",
    "closeness",
    "nl_participants",
    "enzyme_ids"
}
# For some reason integers are not JSON serialisable
# by jinja => casting to strings solves the issue
TO_STRING_ATTRIBUTES = {
    "chain_length",
    "desaturation",
    "hydroxylation",
    "c_index",
    "db_index",
    "oh_index"
}
ATTR_NAMES = {
    # dynamic attributes
    "fold_changes": "Fold Changes",
    "nlog_pvalues": "-log10(FDR)",
    "correlation_changes": "Correlation Changes",
    "partial_correlation_changes": "Partial Correlation Changes",
    "correlations": "Correlations",
    "partial_correlations": "Partial Correlations",
    # static attributes
    "lipid_class": "Lipid Class",
    "reaction_type": "Reaction Type",
    "desaturation": "Desaturation",
    "db_index": "DB Index",
    "chain_length": "Chain Length",
    "c_index": "C Index",
    "hydroxylation": "Hydroxylation",
    "oh_index": "OH Index",
    "degree": "Degree",
    "betweenness": "Betweenness Centrality",
    "closeness": "Closeness Centrality",
    "enzyme_gene_name": "Gene Name",
    "enzyme_uniprot": "Uniprot IDs"
}
NODE_TYPE_SHAPES = {
    "Lipid": "dot",
    "Enzyme": "triangle",
    "Reaction": "triangle",
    "Metabolite": "square"
}


def _setup_scm_(values, n_steps):
    scm = ScalarMappable(
        norm=Normalize(vmin=values["min"],
                       vmax=values["max"]),
        cmap=values["cmap"]
    )
    vals = _continuous_steps_(values, n_steps)
    return scm, vals


def _continuous_steps_(
        colour_map: Dict[str, float],
        n_steps: int
) -> Dict[int, float]:
    legend = {0: colour_map["min"]}
    if colour_map["min"] == colour_map["max"]:
        if colour_map.get("nan") is not None:
            legend[1] = np.nan
        return legend
    step = (colour_map["max"] - colour_map["min"]) / (n_steps - 1)
    for i in range(1, n_steps - 1):
        legend[i] = legend[0] + (i * step)
    legend[n_steps - 1] = colour_map["max"]
    if colour_map.get("nan") is not None:
        legend[n_steps] = np.nan
    return legend


def _size_steps_(
        size_map: Dict[str, Tuple[float, float]],
        n_steps: int
) -> Dict[int, Tuple[float, float]]:
    legend = dict()
    legend[0] = size_map["min"]
    raw_step = (size_map["max"][0] - size_map["min"][0]) / (n_steps - 1)
    norm_step = (size_map["max"][1] - size_map["min"][1]) / (n_steps - 1)
    for i in range(1, n_steps - 1):
        legend[i] = (legend[0][0] + (i * raw_step), legend[0][1] + (i * norm_step))
    legend[n_steps - 1] = size_map["max"]
    if size_map.get("nan") is not None:
        legend[n_steps] = (np.nan, size_map["min"][1]/2)
    return legend


def _node_annotation_(node: dict):
    annotation = f"<strong>{node['label']}</strong><br><br>"
    is_lipid = node['node_molecule_type'] == "Lipid"
    if is_lipid:
        props = STATIC_NODE_PROPERTIES
    else:
        props = STATIC_NON_LIPID_PROPERTIES
    for prop in props:
        node_prop = node.get(prop)
        if node_prop is not None:
            if isinstance(node_prop, float):
                annotation += f"<b>{prop}</b>: {round(node_prop, ndigits=3)}<br>"
            else:
                annotation += f"<b>{prop}</b>: {node_prop}<br>"
    if is_lipid:
        annotation += f"<b>Original Name</b>: {node.get('data_name')}<br>"
    return annotation


def _edge_annotation_(edge: dict):
    annotation = ""
    for prop in STATIC_EDGE_PROPERTIES:
        edge_prop = edge.get(prop)
        if edge_prop is not None:
            if isinstance(edge_prop, float):
                annotation += f"<b>{prop}</b>: {round(edge_prop, ndigits=3)}<br>"
            else:
                annotation += f"<b>{prop}</b>: {edge_prop}<br>"
    return annotation


def _size_filter_(
        sizes: dict,
        n_steps: int
) -> Dict[float, float]:
    keys = np.array(list(sizes.keys()))
    key_dict = _continuous_steps_(
        {"min": np.min(keys),
         "max": np.max(keys)},
        n_steps=n_steps
    )
    vals = np.array(list(sizes.values()))
    val_dict = _continuous_steps_(
        {"min": np.min(vals),
         "max": np.max(vals)},
        n_steps=n_steps
    )
    return dict(zip(key_dict.values(), val_dict.values()))


class VisParser(Network):
    def __init__(self, template: Union[str, Template] = None,
                 **kwargs):
        super(VisParser, self).__init__(**kwargs)
        if template is not None:
            if isinstance(template, str):
                template = Template(open(template, "r").read())
            elif not isinstance(template, Template):
                raise ValueError(
                    "'template' must be a jinja2.Template object or string"
                )
        self.template = template
        self.legend = {"nodes": [], "edges": []}
        self._legend_size_ = 0
        # helper to generate multi column legend
        self._colour_legend_nodes_ = 5
        self._node_cols_ = 1
        self._node_rows_ = 8
        self._to_last_column_ = True

    @classmethod
    def from_pyvis_network(cls, network: Network,
                           template: Template = None,
                           notebook: bool = False) -> VisParser:
        vp = VisParser(template=template,
                       height=network.height,
                       width=network.width,
                       directed=network.directed,
                       bgcolor=network.bgcolor,
                       font_color=network.font_color)
        if notebook:
            vp.prep_notebook()
        return vp

    def _node_update_step_(
            self, attr, value,
            name, counter, colours_to_hex,
            scm, values, cont
    ):
        if attr == "size":
            self.legend["nodes"].append(
                {
                    attr: value,
                    "label": f"{round(name, ndigits=3)}",
                    "id": counter,
                    "x": 0,
                    "y": counter * 2,
                    "shape": "dot",
                    "hidden": False,
                    "fixed": True,
                    "physics": False,
                    "margin": {"top": 0, "right": 15,
                               "bottom": 0, "left": 0},
                }
            )
        elif attr == "color":
            if colours_to_hex:
                if cont:
                    colour = to_hex(scm.to_rgba(value))
                    name_ = f"{round(values[name], ndigits=3)}"
                    # name_ = "{:.3f}".format(values[name])
                else:
                    colour = to_hex(value)
                    name_ = str(name)
            else:
                colour = value
                name_ = str(name)
            self.legend["nodes"].append(
                {
                    "color": colour,
                    "label": name_,
                    "id": counter,
                    "x": 0,
                    "y": counter * 2,
                    "shape": "dot",
                    "hidden": False,
                    "fixed": True,
                    "physics": False,
                    "margin": {"top": 0, "right": 15,
                               "bottom": 0, "left": 0},
                }
            )
        else:
            self.legend["nodes"].append(
                {
                    attr: value,
                    "size": 40,
                    "label": name,
                    "id": counter,
                    "x": 0,
                    "y": counter * 2,
                    "shape": "dot",
                    "hidden": False,
                    "fixed": True,
                    "physics": False,
                    "margin": {"top": 0, "right": 15,
                               "bottom": 0, "left": 0},
                }
            )

    def _update_node_legend_(self, values, attr, i, colours_to_hex,
                             n_steps: int = 5):
        counter = i
        cont = False
        if "cmap" in values.keys():
            scm, values = _setup_scm_(values, n_steps)
            cont = True
        else:
            scm = None
        for name, value in values.items():
            self._node_update_step_(
                attr, value, name, counter,
                colours_to_hex, scm, values, cont
            )
            counter += 1
        return counter

    @staticmethod
    def _dummy_node_(xi, x, y):
        return {
            "id": xi,
            "label": f"dummy{xi}",
            "x": x,
            "y": y,
            "hidden": True,
            "fixed": True,
            "physics": False
        }

    def _update_edge_legend_(self, values, attr, i, colours_to_hex,
                             n_steps: int = 5):
        counter = i
        cont = False
        if "cmap" in values.keys():
            scm, values = _setup_scm_(values, n_steps)
            cont = True
        for name, value in values.items():
            xi = len(self.legend["nodes"])
            self.legend["nodes"].append(
                {
                    "id": xi,
                    "label": f"dummy{xi}",
                    "x": 1,
                    "y": counter * 2,
                    "hidden": True,
                    "fixed": True,
                    "physics": False,
                }
            )
            self.legend["nodes"].append(
                {
                    "id": xi + 1,
                    "label": f"dummy{xi + 1}",
                    "x": 2,
                    "y": counter * 2,
                    "hidden": True,
                    "fixed": True,
                    "physics": False
                }
            )
            if attr == "width":
                self.legend["edges"].append(
                    {
                        "value": value,
                        # "label": "{:.3f}".format(name),
                        "label": f"{round(name, ndigits=3)}",
                        "from": xi,
                        "to": xi + 1,
                        "hidden": False,
                        "fixed": True,
                        "physics": False,
                        "font": {
                            "align": "bottom",
                            "vadjust": 5
                        }
                    }
                )
            elif attr == "color":
                if colours_to_hex:
                    if cont:
                        colour = to_hex(scm.to_rgba(value))
                        # name_ = "{:.3f}".format(values[name])
                        name_ = f"{round(values[name], ndigits=3)}"
                    else:
                        colour = to_hex(value)
                        name_ = str(name)
                else:
                    colour = value
                    name_ = str(name)
                self.legend["edges"].append(
                    {
                        "color": colour,
                        "label": name_,
                        "from": xi,
                        "to": xi + 1,
                        "value": 1,
                        "hidden": False,
                        "fixed": True,
                        "physics": False,
                        "font": {
                            "align": "bottom",
                            "vadjust": 5
                        }
                    }
                )
            else:
                self.legend["edges"].append(
                    {
                        attr: value,
                        "label": name,
                        "from": xi,
                        "to": xi + 1,
                        "value": 1,
                        "hidden": False,
                        "fixed": True,
                        "physics": False,
                        "font": {
                            "align": "bottom",
                            "vadjust": 5
                        }
                    }
                )
            counter += 1
        return counter

    def generate_legend(
        self,
        node_colours: dict = None,
        node_sizes: dict = None,
        node_shapes: dict = None,
        edge_colours: dict = None,
        edge_sizes: dict = None,
        colours_to_hex: bool = True
    ):
        # TODO: handling overwriting to emtpy
        self.legend = {"nodes": [], "edges": []}
        # nodes
        i = 0
        if node_colours is not None:
            i = self._update_node_legend_(node_colours, "color", i,
                                          colours_to_hex)
        if node_sizes is not None:
            # warn("node sizes are not yet supported in legends!")
            i = self._update_node_legend_(_size_filter_(node_sizes, 5),
                                          "size", i, colours_to_hex)
        if node_shapes is not None:
            i = self._update_node_legend_(node_shapes, "shape", i,
                                          colours_to_hex)
        n_nodes = i
        # edges
        i = 0
        if edge_colours is not None:
            i = self._update_edge_legend_(edge_colours, "color", i,
                                          colours_to_hex)
        if edge_sizes is not None:
            warn("edge sizes are not yet supported in legends!")
            i = self._update_edge_legend_(_size_filter_(edge_sizes, 5),
                                          "width", i,
                                          colours_to_hex)
        n_edges = i
        self._legend_size_ = n_nodes if n_nodes > n_edges else n_edges
        # TODO: check if it's possible to increase legend font size

    def _prepare_write_(
        self, template: Template = None,
        force_hierarchical: bool = False,
        use_molecular_strings: bool = False
    ):
        if self.legend is None:
            raise ValueError(
                "Please compute a legend before you call write_vis!"
            )
        if template is None:
            if self.template is None:
                template_path = os.path.join(
                    os.path.dirname(__file__),
                    "templates/vis_main.html"
                )
                template = Template(open(template_path, "r").read())
            else:
                if isinstance(self.template, str):
                    template = Template(open(self.template, "r").read())
                elif isinstance(self.template, Template):
                    template = self.template
                else:
                    raise ValueError(
                        "'template' must be a path or a jinja2 Template object"
                    )
        elif isinstance(template, str):
            template = Template(open(template, "r").read())
        elif not isinstance(template, Template):
            raise ValueError(
                "'template' must be a path or a jinja2 Template object"
            )
            # write network => same as pyvis - should do the trick
            # TODO: notebook compatibility
        use_link_template = True
        # for n in self.nodes:
        #     title = n.get("title", None)
        #     if title:
        #         if "href" in title:
        #             """
        #             this tells the template to override default hover
        #             mechanic, as the tooltip would move with the mouse
        #             cursor which made interacting with hover data useless.
        #             """
        #             use_link_template = True
        #             break
        nodes, edges, heading, height, width, options = self.get_network_data()

        for i in range(len(nodes)):
            for attr in ["degree", "betweenness", "closeness"]:
                nodes[i][attr] = f"{round(nodes[i][attr], ndigits=3)}"
            # adding node_molecule_type based shapes
            nodes[i]['shape'] = NODE_TYPE_SHAPES.get(
                nodes[i].get('node_molecule_type'), 'hexagon'
            )
            if nodes[i].get('node_molecule_type') != 'Lipid':
                nodes[i]['label'] = nodes[i].get('representation_id', '')
            else:
                if use_molecular_strings:
                    mol_string = nodes[i].get('molecular_string')
                    if mol_string is not None:
                        nodes[i]['label'] = nodes[i]['molecular_string']
                    else:
                        nodes[i]['label'] = nodes[i]['data_name']
                else:
                    nodes[i]['label'] = nodes[i]['data_name']
            # if 'inferred' in nodes[i].keys():
            #     del nodes[i]['inferred']
        # check if physics is enabled
        if isinstance(self.options, dict):
            if 'physics' in self.options and 'enabled' in self.options['physics']:
                physics_enabled = self.options['physics']['enabled']
            else:
                physics_enabled = True
        else:
            physics_enabled = self.options.physics.enabled
        # TODO: add more flexibility to this?
        if len(self.edges) > 1200 or force_hierarchical:
            self.options["physics"].solver = "hierarchicalRepulsion"

        return {
            "nodes": nodes,
            "edges": edges,
            "heading": heading,
            "height": height,
            "width": width,
            "options": options,
            "physics_enabled": physics_enabled,
            "template": template,
            "use_link_template": use_link_template
        }

    def write_vis(
        self, name: str, template: Template = None,
        physics_options: dict = None,
        force_hierarchical: bool = False
    ):
        """
        Saving VisParser.network as a html file

        Parameters
        ----------
        name : str
            File name (full path) to write to

        template : jinja2.Template, optional
            Path to a custom html template to use.
            Must be compatible with pyvis and the template
            in '<>/templates/vis_main.html'.

        physics_options : optional
            physics options to pass to pyvis.Network

        force_hierarchical : bool, optional, default False
            Whether to use hierarchical repulsion as a solver
            for network layout. Always True when the number
            of edges exceeds 1200 for performance reasons.

        Returns
        -------
        None
        """
        network_params = self._prepare_write_(template, force_hierarchical)
        self.html = network_params["template"].render(
            nodes=network_params["nodes"],
            edges=network_params["edges"],
            options=network_params["options"],
            physics_enabled=network_params["physics_enabled"],
            use_DOT=self.use_DOT,
            dot_lang=self.dot_lang,
            widget=self.widget,
            bgcolor=self.bgcolor,
            conf=self.conf,
            tooltip_link=network_params["use_link_template"],
            legend_nodes=self.legend["nodes"],
            legend_edges=self.legend["edges"]
        )

        with open(name, "w+") as out:
            out.write(self.html)

    def show(self, name, **kwargs):
        """
        Showing the network in VisParser.network after saving as html

        Parameters
        ----------
        name : str
            File name (full path) to write to

        kwargs :
            Keyword arguments to pass to write_vis.
            Available options:
            template : jinja2.Template, optional
                Path to a custom html template to use.
                Must be compatible with pyvis and the template
                in '<>/templates/vis_main.html'.

            physics_options : optional
                physics options to pass to pyvis.Network

            force_hierarchical : bool, optional, default False
                Whether to use hierarchical repulsion as a solver
                for network layout. Always True when the number
                of edges exceeds 1200 for performance reasons.

        Returns
        -------
        None
        """
        self.write_vis(name, **kwargs)
        webbrowser.open(name)


class DynamicVisParser(VisParser):
    def __init__(
        self, template: Union[str, Template] = None,
        default_colour: str = "#0065bd",
        default_size: int = 3,
        random_seed: int = None, **kwargs
    ):
        if template is None:
            template = os.path.join(
                os.path.dirname(__file__),
                "templates", "vis_dynamic.html"
            )
        # TODO: change template to vis_dynamic.html
        super(DynamicVisParser, self).__init__(
            template=template, **kwargs
        )
        if random_seed is not None:
            layout = self.options.__dict__.get("layout", {})
            layout["randomSeed"] = random_seed
            self.options.__dict__["layout"] = layout
        self._default_colour_ = default_colour
        self._default_size_ = default_size

        self.legend = {
            "nodes": {"colour": [], "size": []},
            "edges": {"colour": [], "size": []},
        }
        # these are just helpers for ensuring visjs node id
        # consecutiveness requirement
        self._edges_by_node_ = {"colour": {}, "size": {}}
        self._legend_attributes_ = dict()
        self._visualisation_options_ = {
            "node_colours": [],
            "node_sizes": [],
            "edge_colours": []
        }

    @classmethod
    def from_pyvis_network(
        cls, network: Network,
        template: Template = None,
        notebook: bool = False
    ) -> DynamicVisParser:
        vp = DynamicVisParser(
            template=template,
            height=network.height,
            width=network.width,
            directed=network.directed,
            bgcolor=network.bgcolor,
            font_color=network.font_color
        )
        if notebook:
            vp.prep_notebook()
        return vp

    @classmethod
    def from_json(
        cls, network_file: str,
        legend_file: str, **kwargs
    ):
        vp = DynamicVisParser(**kwargs)
        # network
        network_data = json.load(open(network_file, "r"))
        vp.nodes = network_data["nodes"]
        vp.edges = network_data["edges"]
        # legend
        vp.legend = json.load(open(legend_file, "r"))
        return vp

    def _labels_to_ids_(self):
        """
        Auxiliary Function to modify edge and node label ids
        to be consecutive integers (otherwise this might lead to
        issues with vis). Also some node properties are explicitly
        set.
        """
        label_to_id = {}
        for i in range(len(self.nodes)):
            node = self.nodes[i]
            label_to_id[node["id"]] = i

            node["color"] = self._default_colour_
            node["physics"] = True
            node["id"] = i
            for attr in node.keys():
                if attr in TO_STRING_ATTRIBUTES:
                    node[attr] = str(node[attr])
            self.nodes[i] = node

        for j in range(len(self.edges)):
            edge = self.edges[j]
            edge["color"] = self._default_colour_
            edge["width"] = self._default_size_
            edge["from"] = label_to_id[edge["from"]]
            edge["to"] = label_to_id[edge["to"]]
            edge["id"] = j
            for attr in edge.keys():
                if attr in TO_STRING_ATTRIBUTES:
                    edge[attr] = str(edge[attr])
            self.edges[j] = edge

    def _update_node_colour_legend_(
        self, name: str, colours: Dict[Any, str],
        n_steps: int, colours_to_hex: bool
    ):
        n = len(self.legend["nodes"]["colour"])
        if name == "lipid_class_colour":
            nlc = len(colours)
            if nlc > 8:
                if nlc > 16:
                    self._node_cols_ = 3
                    if nlc > 24:
                        self._node_rows_ = nlc//3 + 1
                else:
                    self._node_cols_ = 2
                self._colour_legend_nodes_ = nlc
            elif nlc > 5:
                self._colour_legend_nodes_ = nlc
            self._to_last_column_ = 0 < self._colour_legend_nodes_ % self._node_rows_ >= 5

        cont = False
        if "cmap" in colours.keys():
            scm, colours = _setup_scm_(colours, n_steps)
            cont = True
        else:
            scm = None
        for i, (value, colour) in enumerate(colours.items()):
            if colours_to_hex:
                if cont:
                    colour = to_hex(scm.to_rgba(colour))
                else:
                    colour = to_hex(colour)
            if cont:
                # name_ = "{:.3f}".format(colours[i])
                name_ = f"{round(colours[i], ndigits=3)}"
            else:
                name_ = str(value)
            # NOTE: lipid_class comes first if present
            # therefore we do not need to check here!
            x = i // self._node_rows_
            y = i % self._node_rows_
            if i >= n:
                self.legend["nodes"]["colour"].append(
                    {
                        "id": i,
                        "label": "",
                        f"{name}_colour": to_hex(colour),
                        f"{name}_colour_label": name_,
                        "x": x,
                        "y": y * 2,
                        "shape": "dot",
                        "fixed": True,
                        "physics": False,
                        "color": self._default_colour_
                    }
                )
            else:
                elem = {
                    f"{name}_colour": to_hex(colour),
                    f"{name}_colour_label": name_
                }
                ui = (self._node_cols_ - 1) * self._node_rows_ + i
                if self._to_last_column_ or ui < self._colour_legend_nodes_:
                    self.legend["nodes"]["colour"][ui].update(elem)
                else:
                    self.legend["nodes"]["colour"].append(
                        {
                            "id": i,
                            "label": "",
                            f"{name}_colour": to_hex(colour),
                            f"{name}_colour_label": name_,
                            "x": self._node_cols_ - 1,
                            "y": i * 2,
                            "shape": "dot",
                            "fixed": True,
                            "physics": False,
                            "color": self._default_colour_
                        }
                    )

    def _update_node_size_legend_(
        self, name: str, sizes: Dict[str, Tuple[float, float]],
        n_steps: int = 5
    ):
        n = len(self.legend["nodes"]["size"])
        size_map = _size_steps_(sizes, n_steps)
        for i, (value, size) in size_map.items():
            if name == "fold_changes":
                size = abs(size)
            if i >= n:
                self.legend["nodes"]["size"].append(
                    {
                        "id": i,
                        "label": "",
                        f"{name}_size": size,
                        f"{name}_size_label": f"{round(value, ndigits=3)}",
                        "x": 1,
                        "y": i * 2,
                        "shape": "dot",
                        "fixed": True,
                        "physics": False,
                        "color": self._default_colour_
                    }
                )
            else:
                elem = {
                    f"{name}_size": size,
                    f"{name}_size_label": f"{round(value, ndigits=3)}"
                }
                self.legend["nodes"]["size"][i].update(elem)

    def _update_edge_colour_legend_(
        self, name: str, colours: Dict[Any, str],
        n_steps: int, colours_to_hex: bool
    ):
        n = len(self.legend["edges"]["colour"])
        cont = False
        if "cmap" in colours.keys():
            scm, colours = _setup_scm_(colours, n_steps)
            cont = True
        for i, (value, colour) in enumerate(colours.items()):
            xi = (2 * i) + 100
            if colours_to_hex:
                if cont:
                    colour = to_hex(scm.to_rgba(colour))
                else:
                    colour = to_hex(colour)
            if cont:
                name_ = f"{round(colours[i], ndigits=3)}"
            else:
                name_ = str(value)
            if i >= n:
                self.legend["nodes"]["colour"].append(
                    self._dummy_node_(xi, self._node_cols_, i * 2)
                )

                self.legend["nodes"]["colour"].append(
                    self._dummy_node_(xi + 1, self._node_cols_ + 1, i * 2)
                )
                self.legend["edges"]["colour"].append(
                    {
                        f"{name}_colour": colour,
                        f"{name}_colour_label": name_,
                        "from": xi,
                        "to": xi + 1,
                        "id": i,
                        "value": 1,
                        "hidden": False,
                        "fixed": True,
                        "physics": False,
                        "font": {
                            "align": "bottom",
                            "vadjust": 5
                        },
                        "color": self._default_colour_
                    }
                )
                self._edges_by_node_["colour"].setdefault(xi, []).append(i)
                self._edges_by_node_["colour"].setdefault(xi + 1, []).append(i)
            else:
                elem = {
                    f"{name}_colour": colour,
                    f"{name}_colour_label": name_
                }
                self.legend["edges"]["colour"][i].update(elem)

    def _update_edge_size_legend_(
        self, name: str, sizes: Dict[str, Tuple[float, float]],
        n_steps: int = 5
    ):
        n = len(self.legend["edges"]["size"])
        size_map = _size_steps_(sizes, n_steps)
        for i, (value, size) in enumerate(size_map.items()):
            xi = i + 100
            if i >= n:
                self.legend["nodes"]["size"].append(
                    self._dummy_node_(xi, 2, i * 2)
                )
                self.legend["nodes"]["size"].append(
                    self._dummy_node_(xi + 1, 3, i * 2)
                )
                self.legend["edges"]["size"].append(
                    {
                        f"{name}_size": size,
                        f"{name}_size_label": f"{round(value, ndigits=3)}",
                        "from": xi,
                        "to": xi + 1,
                        "value": 1,
                        "id": i,
                        "hidden": False,
                        "fixed": True,
                        "physics": False,
                        "font": {
                            "align": "bottom",
                            "vadjust": 5
                        },
                        "color": self._default_colour_
                    }
                )
                self._edges_by_node_["size"].setdefault(xi, []).append(i)
                self._edges_by_node_["size"].setdefault(xi + 1, []).append(i)
            else:
                elem = {
                    f"{name}_colour": size,
                    f"{name}_size_label": f"{round(value, ndigits=3)}"
                }
                self.legend["edges"]["size"][i].update(elem)

    def generate_legend(
        self,
        node_colours: Dict[str, Dict[Any, str]] = None,
        node_sizes: Dict[str, Dict[str, Tuple[float, float]]] = None,
        node_shapes: Dict[str, Dict[Any, str]] = None,
        edge_colours: Dict[str, Dict[Any, str]] = None,
        edge_sizes: Dict[str, Dict[str, Tuple[float, float]]] = None,
        colours_to_hex: bool = True,
        n_steps: int = 5
    ):
        # node legend
        if node_colours is not None:
            # lipid class is the only discrete node colour option
            # and can therefore be the only attribute with more than 5 legend nodes
            # to have proper multi-row legends it is easiest to do this one first
            class_colours = node_colours.pop("lipid_class_colour")
            if class_colours is not None:
                self._update_node_colour_legend_("lipid_class_colour", class_colours,
                                                 n_steps, colours_to_hex)
                self._legend_attributes_.update({"node_colour": "lipid_class_colour"})
                self._visualisation_options_["node_colours"].append("lipid_class_colour")
            for attr, colours_vals in node_colours.items():
                self._update_node_colour_legend_(attr, colours_vals,
                                                 n_steps, colours_to_hex)
                self._legend_attributes_.update({"node_colour": attr})
                self._visualisation_options_["node_colours"].append(attr)
        if node_sizes is not None:
            for attr, sizes_vals in node_sizes.items():
                self._update_node_size_legend_(attr, sizes_vals,
                                               n_steps)
                self._legend_attributes_.update({"node_size": attr})
                self._visualisation_options_["node_sizes"].append(attr)
        if node_shapes is not None:
            # TODO: implement node shape legend
            print("\nNode Shape legend currently not supported\n")
            # self._legend_attributes_.add("shape")
        # edge legend
        if edge_colours is not None:
            for attr, colours_vals in edge_colours.items():
                self._update_edge_colour_legend_(attr, colours_vals,
                                                 n_steps, colours_to_hex)
                self._legend_attributes_.update({"edge_colour": attr})
                self._visualisation_options_["edge_colours"].append(attr)
        if edge_sizes is not None:
            for attr, sizes_vals in edge_sizes.items():
                self._update_edge_size_legend_(attr, sizes_vals,
                                               n_steps)
                self._legend_attributes_.update({"edge_size": attr})
                self._visualisation_options_["edge_sizes"].append(attr)

    def _consecutive_nodes_(self, attr):
        for i in range(1, len(self.legend["nodes"][attr])):
            c_id = self.legend["nodes"][attr][i]["id"]
            if c_id - self.legend["nodes"][attr][i - 1]["id"] != 1:
                self.legend["nodes"][attr][i]["id"] = i
                for edge_id in self._edges_by_node_[attr].get(c_id, []):
                    c_edge: dict = self.legend["edges"][attr][edge_id]
                    if c_edge["from"] == c_id:
                        self.legend["edges"][attr][edge_id]["from"] = i
                    elif c_edge["to"] == c_id:
                        self.legend["edges"][attr][edge_id]["to"] = i

    @staticmethod
    def _add_annotations_(nodes, edges) -> Tuple[list, list]:
        for i in range(1, len(nodes)):
            node = nodes[i]
            node["fixed_annotation"] = _node_annotation_(node)
            nodes[i] = node
        for i in range(1, len(edges)):
            edge = edges[i]
            edge["fixed_annotation"] = _edge_annotation_(edge)
            edges[i] = edge
        return nodes, edges

    def write_vis(
        self, name: str,
        visualisation_options: Dict[str, List[str]],
        template: Union[Template, str] = None,
        physics_options: dict = None,
        force_hierarchical: bool = False,
        groups: Union[List[str], None] = None,
        as_dict: bool = False,
        group_combinations = None,
        use_molecular_strings: bool = False
    ) -> Union[None, dict]:
        """
        Saving DynamicVisParser.network as a html file

        Parameters
        ----------
        name : str
            File name (full path) to write to

        visualisation_options: dict
            visualisation options to show in html
            TODO: describe properly

        template : jinja2.Template, optional
            Path to a custom html template to use.
            Must be compatible with pyvis and the template
            in '<>/templates/vis_main.html'.

        physics_options : optional
            physics options to pass to pyvis.Network

        force_hierarchical : bool, optional, default False
            Whether to use hierarchical repulsion as a solver
            for network layout. Always True when the number
            of edges exceeds 1200 for performance reasons.

        groups : list of str, optional
            groups for comparisons

        as_dict : bool, optional, default False
            do not use!!!

        Returns
        -------
        None
        """
        # NOTE: for some reason vis cannot handle non-consecutive ids
        # => Dummy nodes for edge legend are renamed at the time of saving.
        #    This enables the flexibility to later add other attributes and
        #    save again.
        self._labels_to_ids_()
        colour_legend = any(["colour" in attr
                             for attr in self._legend_attributes_.keys()])
        size_legend = any(["size" in attr
                           for attr in self._legend_attributes_.keys()])
        if colour_legend:
            self._consecutive_nodes_("colour")
        if size_legend:
            self._consecutive_nodes_("size")
        # getting all relevant elements for saving
        network_params = self._prepare_write_(template, force_hierarchical,
                                              use_molecular_strings)
        # adding fixed annotations to nodes and legends
        network_params['nodes'], network_params['edges'] = \
            self._add_annotations_(network_params['nodes'],
                                   network_params['edges'])
        # ensuring compatibility when no group annotation is provided
        groups = [] if groups is None else groups
        if groups is None:
            group_combinations = []
        elif group_combinations is None:
            group_combinations = [_tuple_to_string_(comb)
                                  for comb in combinations(groups, 2)]
        else:
            group_combinations = [_tuple_to_string_(comb)
                                  for comb in group_combinations]

        if as_dict:
            return dict(
                nodes=network_params["nodes"],
                edges=network_params["edges"],
                options=network_params["options"],
                physics_enabled=network_params["physics_enabled"],
                dot_lang=self.dot_lang,
                widget=self.widget,
                bgcolor=self.bgcolor,
                conf=self.conf,
                tooltip_link=network_params["use_link_template"],
                legend_nodes=self.legend["nodes"]["colour"],
                legend_edges=self.legend["edges"]["colour"],
                legend_size_nodes=self.legend["nodes"]["size"],
                legend_size_edges=self.legend["edges"]["size"],
                legend_attributes=self._legend_attributes_,
                colour_legend=colour_legend,
                size_legend=size_legend,
                single_groups=list(groups),
                groups=group_combinations
            )

        self.html = network_params["template"].render(
            nodes=network_params["nodes"],
            edges=network_params["edges"],
            options=network_params["options"],
            physics_enabled=network_params["physics_enabled"],
            use_DOT=self.use_DOT,
            dot_lang=self.dot_lang,
            widget=self.widget,
            bgcolor=self.bgcolor,
            conf=self.conf,
            tooltip_link=network_params["use_link_template"],
            legend_colour_nodes=self.legend["nodes"]["colour"],
            legend_colour_edges=self.legend["edges"]["colour"],
            legend_size_nodes=self.legend["nodes"]["size"],
            legend_size_edges=self.legend["edges"]["size"],
            colour_legend=colour_legend,
            size_legend=size_legend,
            single_groups=list(groups),
            groups=group_combinations,
            node_colours={attr: ATTR_NAMES[attr] for attr in visualisation_options["node_colours"]},
            edge_colours={attr: ATTR_NAMES[attr] for attr in visualisation_options["edge_colours"]},
            node_sizes={attr: ATTR_NAMES[attr] for attr in visualisation_options["node_sizes"]}
        )

        with open(name, "w+") as out:
            out.write(self.html)

    def show(self, name, visualisation_options, **kwargs):
        self.write_vis(name, visualisation_options, **kwargs)
        webbrowser.open(name)

    def network_to_json(self, file: str, **kwargs):
        json.dump(
            {"nodes": self.nodes, "edges": self.edges},
            open(file, "w"),
            **kwargs
        )

    def to_networkx(self) -> Union[nx.Graph, nx.DiGraph]:
        excluded = {"physics", "size", "id", "width", "shape", "color"}
        graph = nx.DiGraph() if self.directed else nx.Graph()
        for node in self.nodes:
            graph.add_node(
                node["label"],
                **{name: node_attr for name, node_attr in node.items()
                    if name not in excluded}
            )
        for edge in self.edges:
            graph.add_edge(
                edge["from"], edge["to"],
                ** {name: edge_attr for name, edge_attr in edge.items()
                    if name not in excluded}
            )
        return graph

    def legend_to_json(self, file: str, **kwargs):
        json.dump(self.legend, open(file, "w"), **kwargs)

    def legend_to_networkx(self, legend: str = "colour") -> nx.Graph:
        if legend != "colour" and legend != "size":
            raise ValueError(
                f"'legend' must be 'colour' or 'size', not {legend}"
            )
        graph = nx.Graph()
        for node in self.legend["nodes"][legend]:
            graph.add_node(
                node["id"],
                **{name: node_attr for name, node_attr in node.items()
                   if name.endswith(legend) or name.endswith("label")}
            )
        for edge in self.legend["edges"][legend]:
            graph.add_edge(
                edge["from"], edge["to"],
                **{name: edge_attr for name, edge_attr in edge.items()
                   if name.endswith(legend) or name.endswith("label")}
            )
        return graph


if __name__ == "__main__":
    import networkx as nx

    g = nx.karate_club_graph()
    colors = {node: "red" if club == "Officer" else "blue"
              for node, club in nx.get_node_attributes(g, "club").items()}
    nx.set_node_attributes(g, colors, "color")

    vip = VisParser()
    vip.from_nx(g)
    vip.generate_legend(node_colours={"Officer": "red", "Mr. Hi": "blue", "Noop": "green"},
                        edge_colours={"freak": "lightgrey"})
    vip.show("test.html")

    vipc = VisParser(template="templates/vis_graph.html")
    vipc.from_nx(g)
    vipc.generate_legend(node_colours={"Officer": "red", "Mr. Hi": "blue", "Noop": "green"},
                         edge_colours={"freak": "lightgrey"})
    vipc.show("test_c.html")
    print()



# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Union, List, Callable, Tuple, Dict, Any
import pandas as pd
import numpy as np
from matplotlib.pyplot import Line2D


def _size_legend_(extrema: tuple, scale: tuple = None,
                  nodes: bool = True, **kwargs) -> List[Line2D]:
    def size_from_scale(step: int, scale: tuple, extrema: tuple):
        inter = (step - scale[0]) / (scale[1] - scale[0])
        return inter * (extrema[1] - extrema[0])

    if scale is None:
        scale_steps = [abs(extrema[1] - extrema[0]) / 3 * i
                       for i in range(1, 5)]
        sizes = scale_steps
    else:
        scale_steps = [abs(scale[1] - scale[0]) / 3 * i
                       for i in range(1, 5)]
        sizes = [size_from_scale(step, scale, extrema)
                 for step in scale_steps]
    if nodes:
        handles = [Line2D([0], [0],
                          markersize=int(ms),
                          marker="o",
                          markerfacecolor="r",
                          label=round(lab, ndigits=4),
                          **kwargs)
                   for ms, lab in zip(scale_steps, sizes)]
    else:
        handles = [Line2D([0], [0], lw=int(lw),
                          color="r",
                          label=round(lab, ndigits=4),
                          **kwargs)
                   for lw, lab in zip(scale_steps, sizes)]
    return handles


def _range_scale_(
        x: Union[np.ndarray, pd.Series],
        a: Union[int, float], b: Union[int, float],
) -> Union[np.ndarray, pd.Series]:
    min_ = x.min()
    max_ = x.max()
    return ((x - min_) / (max_ - min_) * (b - a)) + a


def _pandas_log_(data: Union[pd.DataFrame, pd.Series],
                 log_fun: Callable = np.log10) -> Union[pd.DataFrame, pd.Series]:
    logged = data.copy()
    log_vals = log_fun(logged.values)
    if isinstance(data, pd.DataFrame):
        return pd.DataFrame(log_vals, index=data.index,
                            columns=data.columns)
    elif isinstance(data, pd.Series):
        return pd.Series(log_vals, index=data.index,
                         name=data.name)
    else:
        raise ValueError(
            "_pandas_log_ is only implemented for pd.Series and pd.DataFrame "
            f"not {type(data).__name__}"
        )


def _reaction_enzyme_annotation(src: str, tgt: str, enzyme: str) -> str:
    return f"{src} -> {tgt}: {enzyme}"


def _aggregate_species_(
        data: Union[pd.DataFrame, pd.Series] = None,
        species_axis: str = "columns"
) -> Union[pd.DataFrame, pd.Series]:
    if isinstance(data, pd.DataFrame):
        species = getattr(data, species_axis)
        dupl_mask = species.duplicated()
        duplicates = species[dupl_mask]
        if species_axis == "columns":
            agg_data = data.copy(deep=True).loc[:, ~dupl_mask]
            for dup_idx in duplicates:
                agg_data.loc[:, dup_idx] = data.loc[:, dup_idx].sum(axis=1)
        elif species_axis == "index":
            agg_data = data.copy(deep=True).loc[~dupl_mask, :]
            for dup_idx in duplicates:
                agg_data.loc[dup_idx, :] = data.loc[dup_idx, :].sum(axis=1)
        else:
            raise ValueError(
                "species_axis must be 'columns' or 'index' "
                f"not {species_axis}"
            )
        return agg_data
    elif isinstance(data, pd.Series):
        return data[~data.duplicated()]
    else:
        raise ValueError(
            "Data must be a pandas DataFrame or Series "
            f"not a {type(data).__name__}"
        )


def _tuple_to_string_(tup: Tuple[str, str]) -> str:
    return f"{tup[0]}_{tup[1]}"


def _get_sizes_report_nas_(
    attr: Dict[str, Any], keys, as_abs=False,
    default=None
):
    sizes = {}
    na_keys = []
    for node in keys:
        val = attr.get(node, default)
        if as_abs:
            val = abs(val)
        sizes[node] = val
        if np.isnan(val):
            na_keys.append(node)
    return sizes, na_keys



from .reaction import FAReaction
from .lipid import FA, Lipid, DBPos
from .exceptions import ParsingError, ReferenceLipidError
from .reference import ReferenceLipid
from typing import List, Dict, Tuple, Union
from numpy import ndarray
import os
import warnings
import re

# NOTE: for some reason LipidLynxX changes the
# working directory in ALL of these three cases:
# 1. module import
# 2. Converter object initialisation
# 3. Converter function calls
wdir = os.getcwd()
try:
    from lynx import Converter

    lynx_not_available: bool = False
except ModuleNotFoundError:
    print("Package 'lynx' (LipidLynxX) not available. Lipid name conversions will not be possible.")
    lynx_not_available: bool = True
os.chdir(wdir)


def db_pos_parser(posis: List[str]) -> List[DBPos]:
    return [DBPos(int(x[:-1]), x[-1]) for x in posis]


def parse_fa_reactions(lines: List[str]) -> List[FAReaction]:
    start = False
    list_of_reactions = []
    for line in lines:
        if start and (line[0] == ">"):
            start = False
        # Parsing of reactions
        # Hashtags are comments
        if start and line[0] != "#":
            fs = line.strip().split(",")
            if len(fs) > 1:
                c = 0
                db = 0
                oh = 0
                ether = False
                lcb = False
                r_name = ""
                for param in fs:
                    if len(param.strip().split(":")) == 3:
                        r_name = param.strip().split(":", maxsplit=1)[0]
                        param = param.strip().split(":", maxsplit=1)[1]
                    if param.strip()[:2] == "C:":
                        c = int(param.split(":")[1].strip())
                    if param.strip()[:2] == "DB":
                        db = int(param.split(":")[1].strip())
                    if param.strip()[:2] == "OH":
                        oh = int(param.split(":")[1].strip())
                    if param.strip() == "ether":
                        ether = True
                    if param.strip() == "lcb":
                        lcb = True
                if c != 0 or db != 0 or oh != 0:
                    list_of_reactions.append(FAReaction(c, db, oh,
                                                        valid_for_ether=ether,
                                                        valid_for_lcb=lcb,
                                                        name=r_name))

        if line.strip() == ">Reactions":
            start = True
    return list_of_reactions


def fa_parser(fa: str, convert_plasmogen: bool = False, force_ether_db_pos: bool = False) -> FA:
    # required params
    oh = 0
    lcb = False
    ether = False
    ether_type = ""
    db_pos = None
    ether_conv = False
    modifications = []
    if "_" in fa or "/" in fa:
        raise ParsingError(f"Error in Fatty acid file: Could not parse fatty acid: {fa}")

    if "iso-" in fa.lower():
        modifications.append("iso")
        fa = re.sub(r"ISO-", "", fa.upper())

    # Hydroxylations
    p1 = re.compile(r'-[0-9]+OH')
    if p1.search(fa) is not None:
        oh += 1
        fa = re.sub(r'-[0-9]+OH', "", fa)

    if "m" in fa or "M" in fa:
        fa = fa.replace("m", "")
        fa = fa.replace("M", "")
        if ";" in fa:
            if fa[-1] == "O":
                fa = fa + "2"
            else:
                fa = fa[0:-1] + str(int(fa[-1]) + 1)
        else:
            fa = fa + ";O"

    if "d" in fa or "D" in fa:
        fa = fa.replace("d", "")
        fa = fa.replace("D", "")
        if ";" in fa:
            if fa[-1] == "O":
                fa = fa + "3"
            else:
                fa = fa[0:-1] + str(int(fa[-1]) + 2)
        else:
            fa = fa + ";O2"

    if "t" in fa or "T" in fa:
        fa = fa.replace("t", "")
        fa = fa.replace("T", "")
        if ";" in fa:
            if fa[-1] == "O":
                fa = fa + "4"
            else:
                fa = fa[0:-1] + str(int(fa[-1]) + 3)
        else:
            fa = fa + ";O3"
    try:
        fs = fa.split(":")
        first = fs[0]
        second = fs[1]
        # Front
        if first[0] == "l" or first[0] == "L":
            lcb = True
            c = int(first[1:])
        elif first[:2].upper() in ["P-", "O-"]:
            ether = True
            ether_type = first[:2].upper()
            c = int(first[2:])
        else:
            c = int(first)
        # Remove
        p = re.compile('[0-9]+[ezEZ]+')
        if "(" in second:
            db_pos = p.findall(second)
            if len(db_pos) > 0:
                second = re.sub(r"\(.*\)", "", second)
            else:
                db_pos = None
            if "OH" in second:
                second = re.sub(r"\(.*\)", "", second)
                oh += 1
        # Back

        if ";" in second:
            db = int(second.split(";")[0])
            # Convert plasmogen to normal ether
            if ether and (ether_type == 'P-') and convert_plasmogen:
                ether_type = 'O-'
                db += 1
                ether_conv = True
            oh_dummy = second.split(";")[1]
            if oh_dummy == "O":
                oh += 1
            elif "O" in oh_dummy:
                oh += int(oh_dummy[1:])
            else:
                oh += int(oh_dummy)
        else:
            db = int(second)
            if ether and (ether_type == 'P-') and convert_plasmogen:
                ether_type = 'O-'
                db += 1
                ether_conv = True

        final_db_pos = db_pos_parser(db_pos) if db_pos is not None else None

        if ether_conv:
            if final_db_pos is not None:
                final_db_pos.append(DBPos(1, 'Z'))
            elif force_ether_db_pos:
                final_db_pos = [DBPos(1, 'Z')]

    except FileNotFoundError:
        raise ParsingError(f"Error in Fatty acid file: Could not parse fatty acid: {fa}")

    return FA(c, db, oh=oh, is_long_chain_base=lcb, is_ether_bond=ether, ether_type=ether_type,
              modifications=modifications,
              db_positions=final_db_pos)


def parse_excluded_fa_reactions(lines: List[str]) -> List[List[FA]]:
    start = False
    list_of_reactions = []
    for line in lines:
        if start and (line[0] == ">"):
            start = False
        # Parsing of reactions
        # Hashtags are comments
        if start and line[0] != "#" and len(line.strip()) > 0:
            fs = line.strip().split(",")
            if len(fs) == 2:
                try:
                    list_of_reactions.append([fa_parser(fs[0].strip()), fa_parser(fs[1].strip())])
                except ParsingError:
                    ParsingError(f"Error in Fatty acid file: Could not read excluded reaction: {line.strip()}")
            else:
                print(f"Excluded reaction must be a list of two fatty acids. Ignoring line: {line.strip()}")
        if line.strip() == ">Excluded Reactions":
            start = True
    return list_of_reactions


def parse_fatty_acids(lines: List[str]) -> Dict[str, List[FA]]:
    start = False
    class_parser = False
    class_dict = {}
    list_of_curr_classes = []
    for line in lines:
        if start and (line[0] == ">"):
            start = False

        if line.strip()[0:8] == "Classes:":
            class_parser = True
            tmp_l = line.strip().split(":")[1]
            if len(tmp_l) == 0:
                list_of_curr_classes = ["all"]
                class_dict["all"] = []
            else:
                splits = tmp_l.split(",")
                splits = [x.strip() for x in splits]
                for a in splits:
                    class_dict[a] = []
                list_of_curr_classes = splits

        if start and line[0] != "#" and len(line.strip()) > 0 and not class_parser:
            fs = line.strip()
            if len(list_of_curr_classes) == 0:
                raise ParsingError("Error in Fatty acid file: No Classes defined for fatty acids")
            try:
                for a in list_of_curr_classes:
                    class_dict[a].append(fa_parser(fs))
            except ParsingError:
                ParsingError(f"Error in Fatty acid file: Could not read excluded fatty acid: {fs}")

        if line.strip() == ">Fatty Acids":
            start = True

        class_parser = False

    return class_dict


def save_lipidlynxx_converter(lipidname: str) -> str:
    if lynx_not_available:
        return lipidname
    try:
        orgwd = os.getcwd()
        conv = Converter()
        os.chdir(orgwd)
        outstr = conv.convert(lipidname.replace("_", "/")).dict()['output']
        os.chdir(orgwd)
        if outstr == '':
            return lipidname
        else:
            return outstr
    except:
        return lipidname


def lipid_parser(lipidname: str,
                 reference_lipids: Dict[str, ReferenceLipid],
                 is_ll_compatible: bool = False,
                 org_name: str = "",
                 convert_plasmogen: bool = False,
                 force_ether_db_pos: bool = False) -> Lipid:
    lipidname_new = lipidname.replace("_", "/").replace(" ", "")
    lipidname_new = re.sub(r'\[.+\]', "", lipidname_new)
    if is_ll_compatible:
        ln = lipidname_new.upper()
    else:
        ln = save_lipidlynxx_converter(lipidname_new.strip()).upper()
    # Get Lipid class
    firstsplit = ln.split("(", maxsplit=1)  # Prevent splitting hydroxilation or DB positions for single fatty acids
    lclass = firstsplit[0]
    if len(firstsplit) == 1:  # e.g. Cholesterol
        if not lclass in reference_lipids.keys():
            raise ReferenceLipidError(f"No compatiable lipid reference class found for Lipid: {lipidname}")
        ref_cls = reference_lipids[lclass]  # Matching reference to lipid

        lip = Lipid(ref_cls.get_abbr(),
                    ref_cls.get_headgroup(),
                    is_molecular_species=True,
                    fa_spots=ref_cls.get_potentialfas(),
                    dataname=org_name,
                    fatty_acids=[],
                    backbone=ref_cls.get_backbone(),
                    category=ref_cls.get_category())

        if ref_cls.is_of_class(lip):
            return lip
        else:
            raise ReferenceLipidError(
                f"Lipid: {lipidname} not compatible with reference class for {ref_cls.get_abbr()}.")

    farest = firstsplit[1]
    if lclass in reference_lipids.keys():
        if reference_lipids[lclass].get_ether():
            lclass = reference_lipids[lclass].get_abbr()
            if lclass[-1] == "P":
                farest = "P-" + farest
            if lclass[-1] == "O":
                farest = "O-" + farest
    # Check potential ethers
    tmp_farest = "/" + farest.upper()
    peth = re.findall(r'/O-', tmp_farest)
    oeth = re.findall(r'/O-', tmp_farest)
    # * Not i lclass
    if ("P-" not in lclass or "O-" not in lclass) and (len(peth)>0 or len(oeth)>0):
        lclass = lclass + " O-"
    # * P- in lclass
    elif ("P-" in lclass) and ("P-" not in farest.upper()):
        farest = "P-" + farest
        lclass = lclass.replace("P-", "O-")
    # * Not in farest
    elif ("O-" in lclass) and ("O-" not in farest.upper() or "P-" not in farest.upper()):
        farest = "O-" + farest
    # Check Class in Reference
    if not lclass in reference_lipids.keys():
        raise ReferenceLipidError(f"No compatiable lipid reference class found for Lipid: {lipidname}")
    ref_cls = reference_lipids[lclass]  # Matching reference to lipid
    # Recognize if SS or MS
    fas = farest[:-1].split("/")
    if len(fas) == ref_cls.get_fas() or len([fa_parser(x, convert_plasmogen=convert_plasmogen, force_ether_db_pos=force_ether_db_pos) for x in fas if x != "0:0"]) == ref_cls.get_fas():  # Molecular species
        if ref_cls.get_lcb():  # Currently only works with one ether and one LCB
            fas[0] = "l" + fas[0]
        lip = Lipid(lipid_class=ref_cls.get_abbr(),
                    head_group=ref_cls.get_headgroup(),
                    is_molecular_species=True,
                    fa_spots=ref_cls.get_potentialfas(),
                    dataname=org_name,
                    fatty_acids=[fa_parser(x, convert_plasmogen=convert_plasmogen, force_ether_db_pos=force_ether_db_pos) for x in fas if x != "0:0"],
                    backbone=ref_cls.get_backbone(),
                    category=ref_cls.get_category())
        if not ref_cls.is_of_class(lip):
            check = 0
            for rl in reference_lipids.values():
                if rl.matches_class_properties(lip):
                    lip = Lipid(lipid_class=rl.get_abbr(),
                                head_group=rl.get_headgroup(),
                                is_molecular_species=True,
                                fa_spots=rl.get_potentialfas(),
                                dataname=org_name,
                                fatty_acids=[fa_parser(x, convert_plasmogen=convert_plasmogen, force_ether_db_pos=force_ether_db_pos) for x in fas if x != "0:0"],
                                backbone=rl.get_backbone(),
                                category=rl.get_category())
                    if rl.is_of_class(lip):
                        return lip
            raise ReferenceLipidError(
                f"Lipid: {lipidname} not compatible with reference class for {ref_cls.get_abbr()}.")
        else:
            return lip

    elif len(fas) == 1:  # Sum species TODO: CL with 2 sum species
        sumfa = fa_parser(fas[0], convert_plasmogen=convert_plasmogen, force_ether_db_pos=force_ether_db_pos)
        lip = Lipid(lipid_class=ref_cls.get_abbr(),
                    head_group=ref_cls.get_headgroup(),
                    is_molecular_species=False,
                    fa_spots=ref_cls.get_potentialfas(),
                    dataname=org_name,
                    sum_length=sumfa.length(),  #
                    sum_dbs=sumfa.db_index(),  #
                    sum_ohs=sumfa.hydroxylations(),  #
                    sum_fa_modifications=sumfa.modifications(),
                    number_fas=ref_cls.get_fas(),
                    has_long_chain_base=ref_cls.get_lcb(),
                    has_ether_bond=ref_cls.get_ether(),
                    ether_type=sumfa.get_ether_type(),
                    backbone=ref_cls.get_backbone(),
                    category=ref_cls.get_category())
        if not ref_cls.is_of_class(lip):
            raise ReferenceLipidError(
                f"Lipid: {org_name} not compatible with reference class for {ref_cls.get_abbr()}.")
        else:
            return lip

    else:
        raise ReferenceLipidError(f"Number of fatty acids of lipid: {lipidname} not compatible with reference class.\n"
                                  f"The lipid can either be a sum species or have {ref_cls.get_fas()} fatty acids "
                                  f"as a molecular species.")


def parse_lipid_list(lipid_input: Union[List[str], ndarray],
                     reference_lipids: Dict[str, ReferenceLipid],
                     ll_compatible: bool = False) -> Tuple[Dict[str, List[Lipid]], Dict[str, str], List[str]]:
    # Return order: lipid_dict, converted_names_mapping (org_name -> converted_name), incompatible_lipids
    converted_names_mapping = {}
    incompatible_lipids = []
    lipid_dict = {}

    for org_name in lipid_input:
        if ll_compatible:
            conv_name = org_name
        else:
            conv_name = save_lipidlynxx_converter(org_name)

        try:
            tmp_lipid = lipid_parser(conv_name,
                                     reference_lipids=reference_lipids,
                                     is_ll_compatible=True,
                                     org_name=conv_name)
        except ReferenceLipidError as e:
            incompatible_lipids.append(org_name)
            warnings.warn(f"No compatible lipid class available. Error message: '{str(e)}'")
        except ParsingError as e:
            incompatible_lipids.append(org_name)
            warnings.warn(f"Error in parsing fatty acid. Error message: '{str(e)}'")
        else:  # If parsing was successful
            converted_names_mapping[org_name] = conv_name
            if tmp_lipid.get_lipid_class() in lipid_dict:
                lipid_dict[tmp_lipid.get_lipid_class()].append(tmp_lipid)
            else:
                lipid_dict[tmp_lipid.get_lipid_class()] = [tmp_lipid]

    return lipid_dict, converted_names_mapping, incompatible_lipids



import warnings

from .reference import ReferenceLipid
from .lipid import Lipid, FA
from .reaction import Reaction, FAReaction
from typing import List, Dict, Union, Tuple
from .edgelist import Edgelist, reaction_str, enzyme_str
from .reaction_evaluation import Extender
from .exceptions import MappingError, MolecularSpeciesError
from .utils import reaction_titles, get_class_reaction
import networkx as nx


def lipid_attributes(lipid: Lipid) -> Dict[str, Union[str, int, float]]:
    if lipid.get_number_fas() == 0:
        # avoid zero division
        return {
            "node_molecule_type": "Lipid", "data_name": lipid.get_dataname(), "lipid_class": lipid.get_lipid_class(),
            "chain_length": lipid.sum_length(), "desaturation": lipid.sum_dbs(), "hydroxylation": lipid.sum_ohs(),
            "c_index": 0, "db_index": 0, "oh_index": 0, "inferred": lipid.converted_to_mol_spec(),
            "molecular_string": str(lipid)
        }
    return {
        "node_molecule_type": "Lipid", "data_name": lipid.get_dataname(), "lipid_class": lipid.get_lipid_class(),
        "chain_length": lipid.sum_length(), "desaturation": lipid.sum_dbs(), "hydroxylation": lipid.sum_ohs(),
        "c_index": lipid.sum_length()/lipid.get_number_fas(), "db_index": lipid.sum_dbs()/lipid.get_number_fas(),
        "oh_index": lipid.sum_ohs()/lipid.get_number_fas(), "inferred": lipid.converted_to_mol_spec(),
        "molecular_string": str(lipid)
    }


def non_lipid_attributes(node_type: str, class_reactions=None, raw_str=None):
    # TODO: add nl_pariticpants etc. here to pass to LipidNetwork
    attrs = {
        "node_molecule_type": node_type, "data_name": None, "lipid_class": "",
        "chain_length": 1, "desaturation": 0, "hydroxylation": 0,
        "c_index": 1, "db_index": 0, "oh_index": 0
    }
    if raw_str is not None:
        id_str = raw_str.split(enzyme_str)[1] if enzyme_str in raw_str else raw_str
        reaction = get_class_reaction(id_str, class_reactions)
        if reaction is None:
            warnings.warn(f'No class reaction found for {raw_str}')
        else:
            attrs["representation_id"] = reaction_titles(id_str, class_reaction=reaction)
            attrs["nl_participants"] = '; '.join(reaction.get_nl_participants())
            attrs["enzyme_ids"] = reaction.get_enzyme_id()
    return attrs


class GenerateLipidNetwork:

    __slots__ = ["__reference_lipids",
                 "__lipids",
                 "__class_reactions",
                 "__fa_reactions",
                 "__excluded_fa_reactions",
                 "__edgelist",
                 "__extender",
                 "__class_fatty_acids",
                 "__has_molecular_species",
                 "__ether_conversions",
                 "__failed_MolecularSpecies"]

    def __init__(self, reference_lipids: List[ReferenceLipid],
                 lipids: Dict[str, List[Lipid]],
                 class_reactions: List[Reaction],
                 fa_reactions: List[FAReaction],
                 excluded_fa_reactions: List[List[FA]] = None,
                 class_fatty_acids: Dict[str, List[FA]] = None,
                 ether_conversions: bool = True,
                 confirmed_species_dict: Dict[str, List[Lipid]] = None,
                 allow_molspec_fails: bool = False):

        if confirmed_species_dict is None:
            confirmed_species_dict = dict()

        # Check if all lipids are molecular species and convert them if not.
        self.__has_molecular_species = False
        self.__failed_MolecularSpecies = []

        new_lipids = dict()
        molspec_error_list = []
        for i in lipids.keys():
            new_lipids[i] = list()
            for l in lipids[i]:
                if l.get_dataname() is None or l.get_dataname() == "":
                    warnings.warn(f"Lipid {str(l)} does not have a data Name, which is required for mapping!")
                if l.is_molecular_species():
                    new_lipids[i].append(l)
                    self.__has_molecular_species = True
                else:
                    if class_fatty_acids is None:
                        raise ValueError("Non molecular lipids in the data, but no class fatty acids provided."
                                         "The class_fatty_acids argument cannot be none")
                    else:
                        # Look at confirmed species
                        if i in confirmed_species_dict.keys():
                            # Loop over all confirmed species
                            tmp_ms_list = []
                            for cfs in confirmed_species_dict[i]:
                                if l.sum_species_str() == cfs.sum_species_str():
                                    cfs.set_dataname(l.get_dataname())
                                    cfs.set_converted_to_mol_spec(converted=True)
                                    tmp_ms_list.append(cfs)

                            if len(tmp_ms_list) > 0:
                                new_lipids[i] = new_lipids[i] + tmp_ms_list
                            else:
                                try:
                                    new_lipids[i] = new_lipids[i] + \
                                                    l.get_molecular_species(class_fatty_acids["all"] if
                                                                            l.get_lipid_class() not in class_fatty_acids else
                                                                            class_fatty_acids[l.get_lipid_class()])
                                except MolecularSpeciesError as e:
                                    molspec_error_list.append(str(e).split(" | ")[1])
                        else:
                            try:
                                new_lipids[i] = new_lipids[i] + \
                                                l.get_molecular_species(class_fatty_acids["all"] if
                                                                        l.get_lipid_class() not in class_fatty_acids else
                                                                        class_fatty_acids[l.get_lipid_class()])
                            except MolecularSpeciesError as e:
                                molspec_error_list.append(str(e).split(" | ")[1])
        if len(molspec_error_list) > 0:
            if not allow_molspec_fails:
                raise MolecularSpeciesError("Molecular species could not be inferred for the following lipids.\n"
                                            "Try to add more possible fatty acids in the fatty acid file.\n"
                                            f"{', '.join(molspec_error_list)}")
            else:
                self.__failed_MolecularSpecies = molspec_error_list
                warnings.warn("Molecular species could not be inferred for the following lipids.\n"
                              "Try to add more possible fatty acids in the fatty acid file.\n"
                              f"{', '.join(molspec_error_list)}")

        self.__reference_lipids = reference_lipids
        self.__lipids = new_lipids
        self.__class_reactions = class_reactions
        self.__fa_reactions = fa_reactions
        self.__excluded_fa_reactions = excluded_fa_reactions
        self.__extender = Extender(class_reactions=class_reactions, fatty_acid_reactions=fa_reactions)
        self.__edgelist = Edgelist()
        self.__class_fatty_acids = class_fatty_acids
        self.__ether_conversions = ether_conversions

    def get_lipid_dict(self) -> Dict[str, List[Lipid]]:
        return self.__lipids

    def has_failed_molspec(self) -> bool:
        return len(self.__failed_MolecularSpecies) > 0

    def failed_molspec(self) -> List[str]:
        return self.__failed_MolecularSpecies

    def add_new_lipid(self, new_lipid: Lipid):
        # Convert to molecular soecies if necessary
        new_lipid.set_dataname("no_data_name")
        if new_lipid.is_molecular_species():
            l = [new_lipid]
        else:
            l = new_lipid.get_molecular_species(self.__class_fatty_acids["all"] if
                                                new_lipid.get_lipid_class() not in self.__class_fatty_acids else
                                                self.__class_fatty_acids[new_lipid.get_lipid_class()])

        if new_lipid.get_lipid_class() in self.__lipids:
            self.__lipids[new_lipid.get_lipid_class()] += l
        else:
            if len(l) > 0:
                self.__lipids[new_lipid.get_lipid_class()] = l
        self.__edgelist = Edgelist()

    def compute_edgelist(self,
                         parallel: bool = False,
                         n_cores: int = 4):
        if parallel:
            warnings.warn("Parallel edgelist computing has been temporarily disabled.\n"
                          "Edgelist will be computed on a single thread.")
            self.__edgelist = self.__extender.evaluate_all(self.__lipids,
                                                           self.__excluded_fa_reactions,
                                                           ether_conversions=self.__ether_conversions,
                                                           reference_lipids=self.__reference_lipids)
            # Disabled parallel:
            # self.__edgelist = self.__extender.evaluate_all_parallel(self.__lipids,
            #                                                         n_cores, self.__excluded_fa_reactions)
        else:
            self.__edgelist = self.__extender.evaluate_all(self.__lipids,
                                                           self.__excluded_fa_reactions,
                                                           ether_conversions=self.__ether_conversions,
                                                           reference_lipids=self.__reference_lipids)

    def molecular_species_network(self, filter_duplicates: bool = False,
                                  excluded_reaction_types: List[str] = None,
                                  return_lipid_mapping: bool = False,
                                  bipartite: bool = False,
                                  bipartite_type: str = "reaction",
                                  multi: bool = False) -> Union[nx.Graph, Tuple[nx.Graph, Dict[str, str]]]:
        if len(self.__edgelist) == 0:
            print("no edgelist available yet. Will be computed using default parameters")
            self.compute_edgelist()
        if multi:
            g = nx.MultiGraph()
        else:
            g = nx.Graph()
        ef = self.__edgelist.molecular_species_edgelist(filter_duplicates=filter_duplicates,
                                                        excluded_reaction_types=excluded_reaction_types,
                                                        bipartite=bipartite, bipartite_type=bipartite_type)
        g.add_edges_from(ef)

        node_list = []
        lipid_mapping = {}

        for lcls in self.__lipids:  # Lipid classes
            for lip in self.__lipids[lcls]:  # Lipids of one class
                node_list.append(lip)
                if lip.get_dataname() in lipid_mapping.keys():
                    lipid_mapping[lip.get_dataname()] = lipid_mapping[lip.get_dataname()] + [str(lip)]
                else:
                    lipid_mapping[lip.get_dataname()] = [str(lip)]

        node_list = list(set(node_list))
        node_list = [(str(x), lipid_attributes(x)) for x in node_list]
        tmp_enz_l = []
        if bipartite:
            if bipartite_type == "reaction":
                tmp_enz_l = [n[1] for n in ef if n[1].startswith(reaction_str)]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, non_lipid_attributes("Reaction")) for x in tmp_enz_l]
            else:
                tmp_enz_l = [n[1] for n in ef if n[1].startswith(enzyme_str)]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, non_lipid_attributes("Enzyme", self.__class_reactions, x)) for x in tmp_enz_l]

        g.add_nodes_from(node_list + tmp_enz_l)

        g.remove_edges_from(nx.selfloop_edges(g))
        if return_lipid_mapping:
            return g, lipid_mapping
        else:
            return g

    def sum_species_network(self, filter_duplicates: bool = False,
                            excluded_reaction_types: List[str] = None, force: bool = False,
                            return_lipid_mapping: bool = False, bipartite: bool = False,
                            bipartite_type: str = "reaction") -> nx.Graph:
        if not force:
            raise MappingError("Sum species network not supported anymore."
                               "The network can be enforced with the parameter force=True.")
        if len(self.__edgelist) == 0:
            print("no edgelist available yet. Will be computed using default parameters")
            self.compute_edgelist()
        g = nx.Graph()
        ef = self.__edgelist.sum_species_edgelist(filter_duplicates=filter_duplicates,
                                                  excluded_reaction_types=excluded_reaction_types,
                                                  bipartite=bipartite, bipartite_type=bipartite_type)
        g.add_edges_from(ef)
        node_list = []
        lipid_mapping = {}

        for lcls in self.__lipids:  # Lipid classes
            for lip in  self.__lipids[lcls]:  # Lipids of one class
                node_list.append((lip.sum_species_str(), {"node_molecule_type": "Lipid"}))
                if lip.get_dataname() in lipid_mapping.keys():
                    lipid_mapping[lip.get_dataname()] = lipid_mapping[lip.get_dataname()] + [lip.sum_species_str()]
                else:
                    lipid_mapping[lip.get_dataname()] = [lip.sum_species_str()]

        node_list = list(set(node_list))

        tmp_enz_l = []
        if bipartite:
            if bipartite_type == "reaction":
                tmp_enz_l = [n[1] for n in ef]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, {"node_molecule_type": "Reaction"}) for x in tmp_enz_l]
            else:
                tmp_enz_l = [n[1] for n in ef]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, {"node_molecule_type": "Enzyme"}) for x in tmp_enz_l]

        tmp_enz_l = list(set(tmp_enz_l))
        g.add_nodes_from(node_list + tmp_enz_l)

        if return_lipid_mapping:
            return g, lipid_mapping
        else:
            return g

    def native_network(self, filter_duplicates: bool = False,
                       excluded_reaction_types: List[str] = None,
                       return_lipid_mapping: bool = False,
                       bipartite: bool = False,
                       bipartite_type: str = "reaction",
                       recompute_edgelist: bool = False,
                       multi: bool = False,
                       verbose=False) -> Union[nx.Graph, Tuple[nx.Graph, Dict[str, str]]]:

        if len(self.__edgelist) == 0 or recompute_edgelist:
            if verbose:
                print("no edgelist available yet. Will be computed using default parameters")
            self.compute_edgelist()
        if multi:
            g = nx.MultiGraph()
        else:
            g = nx.Graph()
        ef = self.__edgelist.native_edgelist(filter_duplicates=filter_duplicates,
                                             excluded_reaction_types=excluded_reaction_types,
                                             bipartite=bipartite, bipartite_type=bipartite_type)
        g.add_edges_from(ef)

        node_list = []
        lipid_mapping = {}

        for lcls in self.__lipids:  # Lipid classes
            for lip in self.__lipids[lcls]:  # Lipids of one class
                node_list.append(lip)
                if lip.get_dataname() in lipid_mapping.keys():
                    if lip.get_native_string() not in lipid_mapping[lip.get_dataname()]:
                        lipid_mapping[lip.get_dataname()] = lipid_mapping[lip.get_dataname()] + [lip.get_native_string()]
                else:
                    lipid_mapping[lip.get_dataname()] = [lip.get_native_string()]

        node_list = list(set(node_list))
        node_list = [(x.get_native_string(), lipid_attributes(x)) for x in node_list]
        tmp_enz_l = []
        if bipartite:
            if bipartite_type == "reaction":
                tmp_enz_l = [n[1] for n in ef if n[1].startswith(reaction_str)]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, non_lipid_attributes("Reaction")) for x in tmp_enz_l]
            else:
                tmp_enz_l = [n[1] for n in ef if n[1].startswith(enzyme_str)]
                tmp_enz_l = list(set(tmp_enz_l))
                tmp_enz_l = [(x, non_lipid_attributes("Enzyme", self.__class_reactions, x)) for x in tmp_enz_l]
        g.add_nodes_from(node_list + tmp_enz_l)
        g.remove_edges_from(nx.selfloop_edges(g))

        if return_lipid_mapping:
            return g, lipid_mapping
        else:
            return g

    def len_edgelist(self) -> int:
        return len(self.__edgelist)

    @property
    def n_lipids(self) -> int:
        return len(self.__lipids)

    def get_class_reactions(self) -> List[Reaction]:
        return self.__class_reactions

    def _get_edgelist(self) -> Edgelist:
        return self.__edgelist



import pandas as pd
import numpy as np
from typing import List, Dict, Tuple
from .reference import ReferenceLipid
from .reaction import Reaction, FAReaction
from .reaction_types import *
from .lipid import Lipid
from .edgelist import Edge, Edgelist
import warnings


def make_organism_reaction_list_from_reactome(molecules: pd.DataFrame,
                                              curated_reactions: pd.DataFrame,
                                              reaction_to_molecules: pd.DataFrame,
                                              ref_lips: Dict[str, ReferenceLipid],
                                              reaction_details: pd.DataFrame,
                                              verbose: bool = True,
                                              organism: str = "HSA") -> List[Reaction]:
    n_reactions = curated_reactions.shape[0]
    merged_reaction_ids = curated_reactions["reaction_IDs"]
    reaction_type = curated_reactions["reaction_type"]
    reaction_notes = curated_reactions["reaction_notes"]
    reaction_details = reaction_details.drop_duplicates().set_index("EnzymeID")
    out_l = []

    organism = organism.upper()

    for r in range(n_reactions):

        subs = []
        prods = []
        if (organism in merged_reaction_ids[r]) and (reaction_type[r] is not np.nan):
            id_l = merged_reaction_ids[r].split(";")

            # Takes only one ID, because we expect the same reaction participants for the same reactions
            tmp_hsa = [x for x in id_l if organism in x]
            hsa_id = tmp_hsa[0]
            hsa_ids = ";".join(tmp_hsa)

            curr_rtm = reaction_to_molecules[reaction_to_molecules["ReactionID"] == hsa_id]
            n_curr_rtm = curr_rtm.shape[0]
            # Non-lipid participants in reactions
            nl_p = []
            for sp in range(n_curr_rtm):
                curr_mol = curr_rtm["MoleculeID"].values[sp]
                curr_part = curr_rtm["Participation"].values[sp]
                curr_st = curr_rtm["Stiochiometry"].values[sp]

                curr_other = molecules[molecules["ID"] == curr_mol]
                if curr_other.shape[0] != 0:
                    if curr_other["Abbreviation"].values[0] is not np.nan:
                        if curr_other["Abbreviation"].values[0].upper() in ref_lips.keys():
                            if curr_part == "Substrate":
                                subs += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                            else:
                                prods += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                        else:
                            nl_p.append(curr_other["name1"].values[0])
                    else:
                        nl_p.append(curr_other["name1"].values[0])

            if len(subs) > 0 or len(prods) > 0:
                if hsa_id in reaction_details.index:
                    tmp_id = reaction_details.loc[hsa_id, :]["UniProtID"]
                    if isinstance(tmp_id, pd.Series):
                        up_id = ", ".join([str(x) for x in tmp_id.values])
                    else:
                        up_id = str(tmp_id)
                    up_id = up_id if up_id != "nan" else ""

                    tmp_id = reaction_details.loc[hsa_id, :]["GeneName"]
                    if isinstance(tmp_id, pd.Series):
                        gn_id = ", ".join([str(x) for x in tmp_id.values])
                    else:
                        gn_id = str(tmp_id)
                    gn_id = gn_id if gn_id != "nan" else ""
                else:
                    up_id = ""
                    gn_id = ""
                new_reac = Reaction(substrates=subs,
                                    products=prods,
                                    reaction_type=reaction_type[r],
                                    enzyme_ids=hsa_ids,
                                    non_lipid_reaction_participants=nl_p,
                                    uniprot=up_id,
                                    gene_name=gn_id)
                if new_reac not in out_l:
                    if verbose:
                        a = [x.get_enzyme_repr() for x in out_l if new_reac.participation_equality(x)
                             and new_reac.get_reaction_type() != x.get_reaction_type()]
                        if len(a) > 0:
                            print("Reaction " + new_reac.get_enzyme_id() + " has the same participants as " + str(a) +
                                  " but different Reaction types.")
                    out_l.append(new_reac)

                else:
                    for tr in range(len(out_l)):
                        if new_reac == out_l[tr]:
                            out_l[tr].extend_enzyme_ids(hsa_ids)
                            if up_id != "":
                                out_l[tr].extend_uniprot(up_id)
                            if gn_id != "":
                                out_l[tr].extend_uniprot(gn_id)
                            out_l[tr].extend_nl_participants(nl_p)
                            break
        else:
            pass

    return out_l


def make_all_reaction_list_from_reactome(molecules: pd.DataFrame,
                                         curated_reactions: pd.DataFrame,
                                         reaction_to_molecules: pd.DataFrame,
                                         ref_lips: Dict[str, ReferenceLipid],
                                         reaction_details: pd.DataFrame,
                                         verbose: bool = True) -> Tuple[List[Reaction], int]:
    n_reactions = curated_reactions.shape[0]
    merged_reaction_ids = curated_reactions["reaction_IDs"]
    reaction_type = curated_reactions["reaction_type"]
    # reaction_notes = curated_reactions["reaction_notes"]
    reaction_details = reaction_details.drop_duplicates().set_index("EnzymeID")
    out_l = []
    conflict_counter = 0
    for r in range(n_reactions):

        if reaction_type[r] is not np.nan:
            id_l = merged_reaction_ids[r].split(";")

            for ids in id_l:
                hsa_id = ids
                subs = []
                prods = []
                curr_rtm = reaction_to_molecules[reaction_to_molecules["ReactionID"] == hsa_id]
                n_curr_rtm = curr_rtm.shape[0]
                nl_p = []
                for sp in range(n_curr_rtm):

                    curr_mol = curr_rtm["MoleculeID"].values[sp]
                    curr_part = curr_rtm["Participation"].values[sp]
                    curr_st = curr_rtm["Stiochiometry"].values[sp]

                    curr_other = molecules[molecules["ID"] == curr_mol]
                    if curr_other.shape[0] != 0:
                        if curr_other["Abbreviation"].values[0] is not np.nan:
                            if curr_other["Abbreviation"].values[0].upper() in ref_lips.keys():
                                if curr_part == "Substrate":
                                    subs += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                                else:
                                    prods += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                            else:
                                nl_p.append(curr_other["name1"].values[0])
                        else:
                            nl_p.append(curr_other["name1"].values[0])
                if len(subs) > 0 or len(prods) > 0:
                    if hsa_id in reaction_details.index:
                        tmp_id = reaction_details.loc[hsa_id, :]["UniProtID"]
                        if isinstance(tmp_id, pd.Series):
                            up_id = ", ".join([str(x) for x in tmp_id.values])
                        else:
                            up_id = str(tmp_id)
                        up_id = up_id if up_id != "nan" else ""

                        tmp_id = reaction_details.loc[hsa_id, :]["GeneName"]
                        if isinstance(tmp_id, pd.Series):
                            gn_id = ", ".join([str(x) for x in tmp_id.values])
                        else:
                            gn_id = str(tmp_id)
                        gn_id = gn_id if gn_id != "nan" else ""

                    else:
                        up_id = ""
                        gn_id = ""
                    new_reac = Reaction(substrates=subs,
                                        products=prods,
                                        reaction_type=reaction_type[r],
                                        enzyme_ids=hsa_id,
                                        non_lipid_reaction_participants=nl_p,
                                        uniprot=up_id,
                                        gene_name=gn_id)
                    if new_reac not in out_l:
                        a = [x.get_enzyme_repr() for x in out_l if new_reac.participation_equality(x)
                             and new_reac.get_reaction_type() != x.get_reaction_type()]
                        if len(a) > 0:
                            l_str = str(a)
                            conflict_counter += 1
                            if verbose:
                                if len(l_str) > 30:
                                    print("Reaction " + new_reac.get_enzyme_id() + " has the same participants as " +
                                          l_str[0:30] + "...] but different Reaction types.\n")
                                else:
                                    print("Reaction " + new_reac.get_enzyme_id() + " has the same participants as " +
                                          l_str + " but different Reaction types.\n")
                        out_l.append(new_reac)

                    else:
                        for tr in range(len(out_l)):
                            if new_reac == out_l[tr]:
                                out_l[tr].extend_enzyme_ids(hsa_id)
                                if up_id != "":
                                    out_l[tr].extend_uniprot(up_id)
                                if gn_id != "":
                                    out_l[tr].extend_uniprot(gn_id)
                                out_l[tr].extend_nl_participants(nl_p)
                                break

        else:
            pass

    return out_l, conflict_counter


def make_all_reaction_list_from_rhea(molecules: pd.DataFrame,
                                     curated_reactions: pd.DataFrame,
                                     reaction_to_molecules: pd.DataFrame,
                                     ref_lips: Dict[str, ReferenceLipid],
                                     gn_mapping: pd.DataFrame,
                                     verbose=True) -> Tuple[List[Reaction], int]:
    n_reactions = curated_reactions.shape[0]
    merged_reaction_ids = curated_reactions["reaction_IDs"]
    reaction_type = curated_reactions["reaction_type"]
    # reaction_notes = curated_reactions["reaction_notes"]

    out_l = []
    conflict_counter = 0
    for r in range(n_reactions):

        if reaction_type[r] is not np.nan:
            id_l = merged_reaction_ids[r].split(";")

            for ids in id_l:
                hsa_id = ids
                subs = []
                prods = []
                curr_rtm = reaction_to_molecules[reaction_to_molecules["ReactionID"] == hsa_id]
                n_curr_rtm = curr_rtm.shape[0]
                nl_p = []
                for sp in range(n_curr_rtm):

                    curr_mol = curr_rtm["MoleculeID"].values[sp]
                    curr_part = curr_rtm["Participation"].values[sp]
                    curr_st = curr_rtm["Stiochiometry"].values[sp]
                    curr_other = molecules[molecules["ChEBI"] == curr_mol]
                    if curr_other.shape[0] != 0:
                        if curr_other["Abbreviation"].values[0] is not np.nan:
                            if curr_other["Abbreviation"].values[0].upper() in ref_lips.keys():
                                if curr_part == "Substrate":
                                    subs += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                                else:
                                    prods += [ref_lips[curr_other["Abbreviation"].values[0].upper()]] * curr_st
                            else:
                                nl_p.append(curr_other["name1"].values[0])
                        else:
                            nl_p.append(curr_other["name1"].values[0])

                if len(subs) > 0 or len(prods) > 0:

                    if hsa_id in gn_mapping.index:
                        tmp_id = gn_mapping.loc[hsa_id, :]["ID"]
                        if isinstance(tmp_id, pd.Series):
                            up_id = ", ".join([str(x) for x in tmp_id.values])
                        else:
                            up_id = str(tmp_id)
                        up_id = up_id if up_id != "nan" else ""

                        tmp_id = gn_mapping.loc[hsa_id, :]["GeneName"]
                        if isinstance(tmp_id, pd.Series):
                            gn_id = ", ".join([str(x) for x in tmp_id.values])
                        else:
                            gn_id = str(tmp_id)
                        gn_id = gn_id if gn_id != "nan" else ""
                    else:
                        up_id = ""
                        gn_id = ""
                    new_reac = Reaction(substrates=subs,
                                        products=prods,
                                        reaction_type=reaction_type[r],
                                        enzyme_ids=hsa_id,
                                        non_lipid_reaction_participants=nl_p,
                                        uniprot=up_id,
                                        gene_name=gn_id)

                    if new_reac not in out_l:
                        a = [x.get_enzyme_id() for x in out_l if new_reac.participation_equality(x)
                             and new_reac.get_reaction_type() != x.get_reaction_type()]
                        if len(a) > 0:
                            l_str = str(a)
                            conflict_counter += 1
                            if verbose:
                                if len(l_str) > 30:
                                    print("Reaction " + new_reac.get_enzyme_id() + " has the same participants as " +
                                          l_str[0:30] + "...] but different Reaction types.\n")
                                else:
                                    print("Reaction " + new_reac.get_enzyme_id() + " has the same participants as " +
                                          l_str + " but different Reaction types.\n")
                        out_l.append(new_reac)

                    else:
                        for tr in range(len(out_l)):
                            if new_reac == out_l[tr]:
                                out_l[tr].extend_enzyme_ids(hsa_id)
                                if up_id != "":
                                    out_l[tr].extend_uniprot(up_id)
                                if gn_id != "":
                                    out_l[tr].extend_uniprot(gn_id)
                                out_l[tr].extend_nl_participants(nl_p)
                                break

        else:
            pass

    return out_l, conflict_counter


def extender(class_reactions: List[Reaction], fa_reactions: List[FAReaction],
             lip_dict: Dict[str, List[Lipid]]) -> Edgelist:
    warnings.warn("extender function in deprecated and will be removed soon. "
                  "Please switch to the Extender class or contact Tim.")
    edge_list = Edgelist()

    # Loop over class reactions
    for i in class_reactions:
        if i.is_valid(verbose=False):
            if i.shape() == (0, 1) or i.shape() == (1, 0):
                pass
            elif i.shape() == (1, 1):
                # Pick both classed from lip_dict
                sub_ref = i.get_substrates()[0]
                prod_ref = i.get_products()[0]
                if sub_ref.get_abbr() in lip_dict and prod_ref.get_abbr() in lip_dict:
                    pot_subs = lip_dict[sub_ref.get_abbr()]
                    pot_prods = lip_dict[prod_ref.get_abbr()]

                    if i.get_reaction_type() == L_HGmodify or i.get_reaction_type() == L_HGdelete:
                        # Loop over both of them
                        for l1 in pot_subs:
                            for l2 in pot_prods:
                                # get pairs
                                if sorted(l1.get_fas()) == sorted(l2.get_fas()):
                                    edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                            reaction_type=i.get_reaction_type(), notes=""))

                    elif i.get_reaction_type() == L_FAdelete:
                        # Loop over both of them
                        for l1 in pot_subs:
                            for l2 in pot_prods:
                                # get pairs
                                f1 = l1.get_fas()
                                f2 = l2.get_fas()

                                for ff1 in range(len(f1)):
                                    if sorted(f1) == sorted(f2+[f1[ff1]]):
                                        edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                                reaction_type=i.get_reaction_type(), notes=""))
                                        break
                                for ff2 in range(len(f2)):
                                    if sorted(f2) == sorted(f1+[f2[ff2]]):
                                        edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                                reaction_type=i.get_reaction_type(), notes=""))
                                        break

            elif i.shape() == (1, 2) or i.shape() == (2, 1):
                if i.get_reaction_type() == L_FAdelete:
                    twos = i.get_substrates() if len(i.get_substrates()) == 2 else i.get_products()
                    ones = i.get_substrates() if len(i.get_substrates()) == 1 else i.get_products()
                    one = ones[0]
                    t1 = twos[0]
                    t2 = twos[1]
                    # find partner
                    if one.get_headgroup() == t1.get_headgroup() and one.get_lcb() == t1.get_lcb() and \
                            one.get_ether() == t1.get_ether():
                        partner = t1
                        other = t2
                    else:
                        partner = t2
                        other = t1
                    if one.get_abbr() in lip_dict and partner.get_abbr() in lip_dict and other.get_abbr() in lip_dict:
                        one_pos = lip_dict[one.get_abbr()]
                        partner_pot = lip_dict[partner.get_abbr()]
                        other_pot = lip_dict[other.get_abbr()]
                        for l1 in one_pos:
                            for l2 in partner_pot:
                                for l3 in other_pot:
                                    for fa in l3.get_fas():
                                        if sorted(l1.get_fas()) == sorted(l2.get_fas() + [fa]):
                                            edge_list.add_edge(
                                                Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                     reaction_type=i.get_reaction_type(), notes=""))
                                            # TODO: Other lipid is skipped for now and not part of the edgelist!

                elif i.get_reaction_type() == L_merge:
                    twos = i.get_substrates() if len(i.get_substrates()) == 2 else i.get_products()
                    ones = i.get_substrates() if len(i.get_substrates()) == 1 else i.get_products()
                    one = ones[0]
                    t1 = twos[0]
                    t2 = twos[1]
                    if one.get_abbr() in lip_dict and t1.get_abbr() in lip_dict and t2.get_abbr() in lip_dict:
                        one_pos = lip_dict[one.get_abbr()]
                        partner_pot = lip_dict[t1.get_abbr()]
                        other_pot = lip_dict[t2.get_abbr()]

                        for l1 in one_pos:
                            for l2 in partner_pot:
                                for l3 in other_pot:
                                    if sorted(l1.get_fas()) == sorted(l2.get_fas() + l3.get_fas()):
                                        edge_list.add_edge(Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                                reaction_type=i.get_reaction_type(), notes=""))
                                        edge_list.add_edge(Edge(l1, l3, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                                reaction_type=i.get_reaction_type(), notes=""))

            elif i.shape() == (2, 2):
                subs = i.get_substrates()
                prods = i.get_products()
                if i.get_reaction_type() == L_FAdelete:
                    if subs[0].get_headgroup() == prods[0].get_headgroup() and \
                       subs[0].get_lcb() == prods[0].get_lcb() and \
                       subs[0].get_ether() == prods[0].get_ether() and \
                       subs[0].get_potentialfas() == prods[0].get_potentialfas():
                        p11 = subs[0]
                        p12 = prods[0]
                        p21 = subs[1]
                        p22 = prods[1]
                    else:
                        p11 = subs[0]
                        p12 = prods[1]
                        p21 = subs[1]
                        p22 = prods[0]

                    if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and p21.get_abbr() in lip_dict and \
                            p22.get_abbr() in lip_dict:
                        for l1 in lip_dict[p11.get_abbr()]:
                            for l2 in lip_dict[p12.get_abbr()]:
                                for l3 in lip_dict[p21.get_abbr()]:
                                    for l4 in lip_dict[p22.get_abbr()]:
                                        fa_set = set(l1.get_fas() + l2.get_fas() + l3.get_fas() + l4.get_fas())
                                        for fa in fa_set:
                                            if (sorted(l1.get_fas()) == sorted(l2.get_fas() + [fa]) or
                                               sorted(l1.get_fas() + [fa]) == sorted(l2.get_fas())) and \
                                               (sorted(l3.get_fas()) == sorted(l4.get_fas() + [fa]) or
                                               sorted(l3.get_fas() + [fa]) == sorted(l4.get_fas())):
                                                edge_list.add_edge(
                                                    Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                         reaction_type=i.get_reaction_type(), notes=""))
                                                edge_list.add_edge(
                                                    Edge(l3, l4, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                         reaction_type=i.get_reaction_type(), notes=""))

                elif i.get_reaction_type() == L_HGdelete or i.get_reaction_type() == L_HGmodify:
                    if subs[0].get_headgroup() == prods[0].get_headgroup() and \
                            subs[0].get_lcb() == prods[1].get_lcb() and subs[0].get_ether() == prods[1].get_ether():
                        p11 = subs[0]
                        p12 = prods[1]
                        p21 = subs[1]
                        p22 = prods[0]
                    else:
                        p11 = subs[0]
                        p12 = prods[0]
                        p21 = subs[1]
                        p22 = prods[1]
                    if p11.get_abbr() in lip_dict and p12.get_abbr() in lip_dict and p21.get_abbr() in lip_dict and \
                            p22.get_abbr() in lip_dict:
                        for l1 in lip_dict[p11.get_abbr()]:
                            for l2 in lip_dict[p12.get_abbr()]:
                                for l3 in lip_dict[p21.get_abbr()]:
                                    for l4 in lip_dict[p22.get_abbr()]:
                                        if sorted(l1.get_fas()) == sorted(l2.get_fas()) and \
                                                sorted(l3.get_fas()) == sorted(l4.get_fas()):
                                            edge_list.add_edge(
                                                Edge(l1, l2, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                     reaction_type=i.get_reaction_type(), notes=""))
                                            edge_list.add_edge(
                                                Edge(l3, l4, enzyme_id=i.get_enzyme_repr(), reaction_id="r_id",
                                                     reaction_type=i.get_reaction_type(), notes=""))

    # Fatty Acid reactions
    for far in fa_reactions:
        # loop over lipids -> FA modifications are only possible for lipids of one class
        for lip in lip_dict:
            if len(lip_dict[lip]) > 1:
                for l1 in lip_dict[lip]:
                    for l2 in lip_dict[lip]:
                        if far.is_transformable(l1, l2):
                            edge_list.add_edge(Edge(l1, l2, enzyme_id="", reaction_id="r_id",
                                                    reaction_type="L_FAmodify", notes=far.get_last_conversion()))

    return edge_list




# This file includes code from:
# LINEX webtool
# Copyright (C) 2021  LipiTUM group, Chair of experimental Bioinformatics, Technical University of Munich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
from typing import Union, Callable, Tuple, Set, List, Dict

import networkx as nx
import numpy as np
import pandas as pd
from scipy import linalg, stats
from sklearn.covariance import (
    GraphicalLasso,
    LedoitWolf,
    empirical_covariance,
    ShrunkCovariance
)
from statsmodels.stats.moment_helpers import cov2corr
from statsmodels.stats.multitest import multipletests
from typing import Iterable, Optional
from .exceptions import CorrelationError, PartialCorrelationError
from .lipid import FA, Lipid
from .reaction import Reaction
import seaborn as sns
import matplotlib.pyplot as plt
import math
import pkg_resources
import json


lipid_colours = json.load(
    open(pkg_resources.resource_filename('linex2', 'templates/lipid_colours.json'), 'r')
)


def partial_correlations(data: pd.DataFrame,
                         estimator: str = "GraphLasso",
                         **kwargs) -> Tuple[pd.DataFrame, np.ndarray]:
    """
    Computing all feature partial correlations. Implementation based on
    Whittaker (1990) and the R package pppcor

    Parameters
    ----------
    :param data: pd.DataFrame, samples in rows
    :param estimator: str, 'empirical', 'GraphLasso', 'Shrinkage' or 'LedoitWolf', covariance estimator
    :return: tuple of pd.Series with partial correlations and pvalues, indices are data.columns
    """
    # computing covariance matrix
    if estimator == "empirical":
        try:
            cov = empirical_covariance(data, **kwargs)
        # TODO: what other errors can occur?
        except ValueError:
            raise PartialCorrelationError(
                f"Invalid values for empirical covariance calculation!",
                estimator
            )
    elif estimator == "GraphLasso":
        # TODO: should we use GraphicalLassoCV here?
        try:
            graph_lasso = GraphicalLasso()
            graph_lasso.fit(data)
            cov = graph_lasso.covariance_
        except ValueError:
            raise PartialCorrelationError(
                f"Invalid values for GraphLasso partial correlation calculations!",
                estimator
            )

    elif estimator == "LedoitWolf":
        try:
            lw = LedoitWolf()
            lw.fit(data)
            cov = lw.covariance_
        except ValueError:
            raise PartialCorrelationError(
                f"Invalid values for Ledoit-Wolf partial correlation calculations!",
                estimator
            )
    elif estimator == "GraphLasso":
        try:
            shrunk = ShrunkCovariance()
            shrunk.fit(data)
            cov = shrunk.covariance_
        except ValueError:
            raise PartialCorrelationError(
                f"Invalid values for GraphLasso partial correlation calculations!",
                estimator
            )
    else:
        raise ValueError(
            f"Estimator must be one of ['GraphLasso', 'Shrinkage', 'LedoitWolf', 'empirical'], not {estimator}"
        )
    # inverse covariance matrix
    try:
        inv_cov = linalg.inv(cov)
    except ValueError:
        raise PartialCorrelationError(
            f"Invalid values in covariance matrix inversion",
            estimator
        )
    # partial correlation matrix
    pcor = -cov2corr(inv_cov)
    np.fill_diagonal(pcor, 1)
    # p-values from fisher's z-transform
    n, m = data.shape
    fact = n - (m - 2) - 3
    pvals = -np.ones(pcor.shape)
    if fact > 0:
        stats_ = np.sqrt(fact) * .5 * np.log((1 + pcor)/(1 - pcor))
        # TODO: double check if normality actually is true
        pvals = 2 * stats.norm.cdf(1 - stats_)
    else:
        print("p-values could not be calculated due to too few samples")

    pcor_df = pd.DataFrame(pcor, index=data.columns,
                           columns=data.columns)
    return pcor_df, pvals


def correlations(data: pd.DataFrame,
                 method: str = "spearman",
                 **kwargs) -> Tuple[pd.DataFrame, np.ndarray]:
    methods = {
        "pearson": stats.pearsonr,
        "spearman": stats.spearmanr,
        "kendall": stats.kendalltau
    }

    if method not in methods.keys():
        raise ValueError(
            f"'method' must be one of {list(methods.keys())}"
        )

    if data.shape[1] < 2:
        cors = pd.DataFrame(columns=data.index,
                            index=data.index)
        pvals = pd.DataFrame(columns=data.index,
                             index=data.index)
        return cors, pvals.values

    cors = np.zeros(2*[data.shape[0]])
    pvals = np.zeros(2*[data.shape[0]])
    for i in range(data.shape[0] - 1):
        for j in range(i + 1, data.shape[0]):
            try:
                cor_, pval_ = methods[method](data.values[i, :],
                                              data.values[j, :],
                                              **kwargs)
            except ValueError:
                if data.iloc[i, :].isna().values.any() or data.iloc[j, :].isna().values.any():
                    cor_ = np.nan
                    pval_ = np.nan
                else:
                    lipids = (data.index[i], data.index[j])
                    raise CorrelationError(
                        f"Invalid values at lipids {lipids} during correlation calculations.\n",
                        lipids
                    )
            cors[i, j] = cor_
            cors[j, i] = cor_
            pvals[i, j] = pval_
            pvals[j, i] = pval_

    cor_df = pd.DataFrame(cors, columns=data.index,
                          index=data.index)
    return cor_df, pvals


def _matrix_pval_correction_(data: Union[pd.DataFrame, np.ndarray],
                             **kwargs) -> np.ndarray:
    """
    Multiple test correction on a symmetric(!) p-value matrix

    Parameters
    ----------
    data
    kwargs

    Returns
    -------

    """
    lidx = np.tril_indices_from(data, -1)

    if isinstance(data, pd.DataFrame):
        pval_corr = multipletests(data.values[lidx], **kwargs)
    else:
        pval_corr = multipletests(data[lidx], **kwargs)

    corrected = np.zeros(data.shape)
    # corrected[uidx] = pval_corr[1]
    corrected[lidx] = pval_corr[1]
    corrected += corrected.T

    return corrected


def generalised_log(data, c=1e-5,
                    log_fun=np.log2):
    return log_fun(data + np.sqrt(data ** 2 + c))


def fold_changes(data: pd.DataFrame,
                 groups: Union[np.ndarray, pd.Series],
                 compare_groups: Union[Tuple[str], List[str]],
                 data_is_log: bool = True,
                 log_func=generalised_log,
                 to_log: bool = False) -> pd.Series:
    """
    Helper function for computing (log-) fold changes of means of samples groups.

    Parameters
    ----------
    data : data from which to compute fold changes. Samples must be in columns
    groups : array of sample groups in the order of data.columns
    compare_groups : list of strings of the two groups to compare
    data_is_log : boolean, indicating whether input data is already log-transformed
                  if False, data will be in-line log-transformed
    log_func : function for log transformation
    to_log: boolean, whether to do log transformation if data_is_log is False

    Returns
    -------
    pd.Series of size data.shape[0] containing fold changes
    """
    group1 = data.loc[:, groups == compare_groups[0]]
    group2 = data.loc[:, groups == compare_groups[1]]
    if data_is_log:
        fcs = np.nanmean(group1, axis=1) - np.nanmean(group2, axis=1)
        return pd.Series(fcs, index=data.index)
    elif to_log:
        fcs = np.nanmean(log_func(group1 + 1), axis=1) - np.nanmean(log_func(group2 + 1), axis=1)
        return pd.Series(fcs, index=data.index)
    else:
        m1 = np.nanmean(group1, axis=1)
        m2 = np.nanmean(group2, axis=1)
        fcs = pd.Series(np.zeros(data.shape[0]),
                        index=data.index)
        for i in range(data.shape[0]):
            if m1[i] >= m2[i]:
                if m2[i] == 0:
                    fcs[i] = np.nan
                else:
                    fcs[i] = m1[i]/m2[i]
            else:
                if m1[i] == 0:
                    fcs[i] = np.nan
                else:
                    fcs[i] = -(m2[i] / m1[i])
        return fcs


def binary_test(data: pd.DataFrame,
                groups: Union[np.ndarray, pd.Series],
                compare_groups: Union[List[str], Set[str], Tuple[str, str]],
                method: Union[str, Callable] = "ttest",
                p_adjust_method: str = "fdr_bh",
                **kwargs) -> pd.Series:
    methods = {
        "wilcoxon": stats.wilcoxon,
        "mannwhitneyu": stats.mannwhitneyu,
        "ranksums": stats.ranksums,
        "ttest": stats.ttest_ind
    }
    if isinstance(method, str):
        test = methods[method]
    else:
        test = method
    # subsetting groups
    if np.unique(groups).size == 2:
        f_groups = groups
    else:
        f_groups = groups[np.isin(groups, compare_groups)]
    # subsetting data to only contain groups of interest
    sub_data = data.loc[:, f_groups.index]
    # computing feature-wise p-values
    pvals = sub_data.apply(lambda x: test(x.values[f_groups == compare_groups[0]],
                                          x.values[f_groups == compare_groups[1]],
                                          **kwargs)[1],
                           axis=1)
    # NOTE: nans cause multipletest to return nans only!
    pvals[pvals.isna()] = 1
    return pd.Series(multipletests(pvals.values, method=p_adjust_method)[1],
                     index=pvals.index)


def unique_elements(data: Union[pd.DataFrame, pd.Series, dict, SymmetricMatrix]) -> np.ndarray:
    """
    Helper to extract all unique elements from a data frame or a series.<br>

    For a pd.Series the result is equivalent to pd.Series.unique().values

    Parameters
    ----------
    data : a general value

    Returns
    -------

    """
    if isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        nas = data.isna()
        na_free = data.values[np.invert(nas)]
    elif isinstance(data, SymmetricMatrix):
        try:
            na_free = data.data[~np.isnan(data.data)]
        # catching numpy being unable to use isnan with string objects
        except TypeError:
            na_free = data.data[~pd.DataFrame(data.data).isna().values]
    else:
        vals = np.array(list(data.values()))
        try:
            na_free = vals[~np.isnan(vals)]
        # catching numpy being unable to use isnan with string objects
        except TypeError:
            na_free = vals[[val != np.nan for val in vals]]
    return np.unique(na_free)


def correlation_change(c_s_: SymmetricMatrix,
                       c_t_: SymmetricMatrix) -> SymmetricMatrix:

    if not c_s_.shape == c_t_.shape:
        raise ValueError(
            "Both input matrices need to have the same input dimensions!"
        )
    c_s_.fillna(0)
    c_t_.fillna(0)
    # make sure c_s_ and c_t_'s indices are ordered the same
    c_t_.align(c_s_)
    x_ = np.tril_indices_from(c_s_.data)
    x = c_s_.data[x_]
    y = c_t_.data[x_]

    inter = np.array(x.size * [""], dtype=object)
    # unsignificant
    inter[(x == 0) & (y == 0)] = "unsignificant"
    # significant vs. unsignificant
    inter[((x != 0) & (y == 0))] = "significant to unsignificant"
    inter[((x == 0) & (y != 0))] = "unsignificant to significant"
    # unchanged significant
    # NOTE: needs to come before the next two or else will overwrite
    inter[(x != 0) & (y != 0)] = "unchanged significant"
    # both significant
    inter[(x > 0) & (y < 0)] = "positive to negative"
    inter[(x < 0) & (y > 0)] = "negative to positive"

    ret = np.array(c_s_.shape[1]*[c_s_.shape[0]*[""]],
                   dtype=object)
    ret[x_] = inter
    ret.T[x_] = inter
    return SymmetricMatrix(
        pd.DataFrame(ret, columns=c_s_.index,
                     index=c_s_.index)
    )


def adapt_data(dat: pd.DataFrame, mapping: Dict[str, str]) -> pd.DataFrame:
    indices = dat.index.values
    fi = []
    newname = []
    for i in range(len(indices)):
        if indices[i] in mapping.keys():
            fi.append(i)
            newname.append(mapping[indices[i]])
    d = dat.copy()
    d = d.iloc[fi, :]
    d.index = pd.Series(newname)

    return d


class SymmetricMatrix:
    __slots__ = [
        "data", "idx_map"
    ]

    data: np.ndarray
    idx_map: Dict[str, int]

    def __init__(self, data: pd.DataFrame = None):
        # TODO: use only an upper triangular matrix => linearize
        if data is not None:
            self.set_data(data)

    def set_data(self, data: pd.DataFrame):
        if np.any(data.index != data.columns):
            raise ValueError(
                "Input data must be a symmetric matrix"
            )
        duplicates = data.index.duplicated()
        if np.any(duplicates):
            dups = data.index[duplicates]
            raise ValueError(
                f"Duplicated indices found in 'SymmetricMatrix' call: "
                f"{', '.join(dups)}"
            )
        self.data = data.values
        self.idx_map = {name: idx for idx, name in enumerate(data.index.values)}

    def _get_idxs_(self, it):
        if isinstance(it, slice):
            if not (it.start is None and it.stop is None and it.step is None):
                raise ValueError("Slicing between index ranges not supported")
            return it
        elif isinstance(it, str):
            return self.idx_map[it]
        elif isinstance(it, Iterable):
            return [self.idx_map[xi] for xi in it]

    def __lt__(self, other):
        return self.data.__lt__(other)

    def __gt__(self, other):
        return self.data.__gt__(other)

    def __eq__(self, other):
        if other is None:
            return self.is_empty()
        elif isinstance(other, SymmetricMatrix):
            return self.data == other.data
        elif isinstance(other, np.ndarray):
            return self.data == other
        else:
            raise ValueError(
                f'No comparison implemented for type {type(other).__name__}.'
            )

    def __ne__(self, other):
        return self.data.__ne__(other)

    def __neg__(self, other):
        return self.data.__neg__(other)

    def __add__(self, other):
        return self.data.__add__(other)

    def __sub__(self, other):
        return self.data.__sub__(other)

    def __getitem__(self, item):
        if isinstance(item, list) or isinstance(item, tuple) or isinstance(item, np.ndarray):
            if len(item) != 2:
                raise ValueError("Only two dimensional slicing supported!")
            return self.data[self._get_idxs_(item[0]), self._get_idxs_(item[1])]
        elif isinstance(item, str):
            return self.data[self.idx_map[item], self.idx_map[item]]
        else:
            raise ValueError(f"{type(item).__name__} not a supported type for slicing")

    def get(self, row_key, col_key, default=None):
        row_idx = self.idx_map.get(row_key)
        col_idx = self.idx_map.get(col_key)
        if row_idx is None or col_idx is None:
            return default
        # since data is a symmetric matrix, the above check
        # is sufficient to guarantee this slice works
        return self.data[row_idx, col_idx]

    @property
    def keys(self):
        return self.idx_map.keys()

    def __repr__(self):
        if hasattr(self, "data"):
            return self.data.__repr__()
        return "[]"

    # to have a dict-like behaviour for if SymmetricMatrix: do...
    def __nonzero__(self):
        return self.is_empty()

    def is_empty(self):
        return not hasattr(self, "data")

    @property
    def shape(self):
        return self.data.shape

    @property
    def index(self):
        idx = np.empty(len(self.idx_map), dtype=object)
        for key, val in self.idx_map.items():
            idx[val] = key
        return idx

    def fillna(self, value, inplace: bool = True) -> Union[None, np.ndarray]:
        nans = np.isnan(self.data)
        if inplace:
            self.data[nans] = value
        else:
            data = self.data.copy()
            data[nans] = value
            return data

    def align(self, matrix: SymmetricMatrix):
        if len(self.idx_map) != len(matrix.idx_map):
            raise ValueError(
                f"'matrix' must not have a different number of entries"
            )
        key_sort = np.empty(len(self.idx_map), dtype=np.int32)
        idx_map_upd = {}
        # TODO: make sure this working as expected (i.e. double check!)
        try:
            for key, val in matrix.idx_map.items():
                key_sort[self.idx_map[key]] = val
                idx_map_upd[key] = val
        except KeyError:
            unmatched = [key for key in matrix.idx_map.keys()
                         if self.idx_map.get(key) is None]
            raise KeyError(
                f"The following keys could not be found: {unmatched}"
            )
        self.data = self.data[key_sort, :][:, key_sort]
        self.idx_map = idx_map_upd

    def to_csv(self, *args, **kwargs):
        indices = np.empty(len(self.idx_map), dtype=object)
        for key, idx in self.idx_map.items():
            indices[idx] = key
        pd.DataFrame(
            self.data, columns=indices, index=indices
        ).to_csv(*args, **kwargs)


def extend_class_fa(class_fa: Dict[str, List[FA]],
                    lipid_dict: Dict[str, List[Lipid]],
                    add_to_all: bool = True) -> Dict[str, List[FA]]:
    """
    Extend class specific fatty acids with known fatty acids from molecular species for that class

    Parameters
    ----------
    :param class_fa: Dict[str, List[FA]]
        Class fatty acids as generated by the function "parse_fatty_acids".

    :param lipid_dict: Dict[str, List[Lipid]]
        Dict of Lipids. With Lipid class as key and list of Lipids as values.

    :param add_to_all: bool, default=True
        Add fatty acids of lipid classes, which are not specified to fatty acids of all unspecified classes.

    Returns
    -------
    :return:
        Extended dict of class specific fatty acids
    """

    # Make copy of class_fa dict
    new_class_fa = {x: class_fa[x].copy() for x in class_fa.keys()}

    # Loop over all classes
    for cls in lipid_dict.keys():

        if cls in new_class_fa.keys():
            add_to = cls
        else:
            if add_to_all:
                add_to = "all"
            else:
                continue

        class_lipids = lipid_dict[cls]
        # Loop over every lipid
        for lip in class_lipids:
            # Do not use fatty acids, if molecular species is inferred from sum species
            if not lip.converted_to_mol_spec():
                if lip.is_molecular_species():
                    for fa in lip.get_fas():
                        if fa not in new_class_fa[add_to]:
                            new_class_fa[add_to].append(fa)

    return new_class_fa


def check_databases(dbs: Union[List[str], Tuple[str]]) -> bool:
    counter = 0
    if "Reactome" in dbs:
        counter += 1
    if "Rhea" in dbs:
        counter += 1

    if (len(dbs) > counter) or (counter < 1):
        raise ValueError("Selected databases must be 'Reactome', 'Rhea' or ('Reactome', 'Rhea').")
    else:
        return True


def combine_reactions(reactions1: List[Reaction], reactions2: List[Reaction]) -> List[Reaction]:

    # copy content from reaction1
    reactions = [reac for reac in reactions1]

    # loop over all other reactions
    for r2 in reactions2:
        found = False
        for r1i in range(len(reactions)):
            if r2.participation_equality(reactions[r1i]):
                reactions[r1i].extend_enzyme_ids(r2.get_enzyme_id())
                reactions[r1i].extend_uniprot(r2.get_uniprot())
                reactions[r1i].extend_gene_name(r2.get_gene_name())
                reactions[r1i].extend_nl_participants(r2.get_nl_participants())
                found = True
                break
        if not found:
            reactions.append(r2)

    return reactions


def plot_boxplot(data: Dict[str, pd.DataFrame],
                 plot_n_col: int = 3,
                 plot_fig_size: Tuple[int, int] = (10, 18),
                 plot_hspace: float = 0.4,
                 difference_dist: bool = False,
                 as_subplots: bool = True,
                 z_scores: bool = True
                 ) -> plt.Figure:
    if as_subplots:
        def plot_boxes(ax, data, key, difference_dist, z_scores):
            if z_scores:
                data["value"] = (data['value'] - data['value'].mean()) / data['value'].std()
            if "groups" in data.columns:
                sns.boxplot(
                    ax=ax,
                    data=data,
                    x="groups",
                    y="value",
                ).set_title(key)
            else:
                sns.boxplot(
                    ax=ax,
                    data=data,
                    y="value",
                ).set_title(key)
            if difference_dist:
                ax.set(ylabel='Ratio difference')
            else:
                ax.set(ylabel='Ratio')
        return plot_template(
            data, plot_boxes, plot_n_col, plot_fig_size,
            plot_hspace, difference_dist=difference_dist,
            z_scores=z_scores
        )
    else:
        fig, ax = plt.subplots(figsize=plot_fig_size)
        dfs = []
        for reaction, df in data.items():
            if z_scores:
                df["value"] = (df['value'] - df['value'].mean()) / df['value'].std()
            df['Reaction'] = reaction
            dfs.append(df)
        df_data = pd.concat(dfs)

        if 'groups' in df_data.columns:
            sns.boxplot(
                ax=ax,
                data=df_data,
                y="value",
                x='Reaction',
                hue='groups'
            )
        else:
            sns.boxplot(
                ax=ax,
                data=df_data,
                y="value",
                x='Reaction'
            )
        if difference_dist:
            ax.set(ylabel='Ratio difference')
        else:
            ax.set(ylabel='Ratio')
        return fig


def plot_histogram(data: Dict[str, pd.DataFrame],
                   plot_n_col: int = 3,
                   plot_fig_size: Tuple[int, int] = (10, 18),
                   plot_hspace: float = 0.4,
                   difference_dist: bool = False
                   ) -> plt.Figure:

    # Seaborn histplot tutorial: https://seaborn.pydata.org/generated/seaborn.histplot.html
    # plotting tutorial: https://realpython.com/python-matplotlib-guide/
    def plot_hist(ax, data, key, difference_dist):
        if "groups" in data.columns:
            sns.histplot(ax=ax,
                         data=data,
                         x="value",
                         hue="groups",
                         element="poly",
                         stat="count"
                         ).set_title(key)
        else:
            sns.histplot(ax=ax,
                         data=data,
                         x="value",
                         element="poly",
                         stat="count"
                         ).set_title(key)
        if difference_dist:
            ax.set(xlabel='Ratio difference', ylabel='Count')
        else:
            ax.set(xlabel='Ratio', ylabel='Count')

    return plot_template(
        data, plot_hist, plot_n_col, plot_fig_size,
        plot_hspace, difference_dist=difference_dist
    )


def get_class_reaction(raw_id: str, cls_reacs: List[Reaction]) -> Union[None, Reaction]:
    if "\nFA: " in raw_id:
        tmp_id = raw_id.split("\nFA: ")[0]
    else:
        tmp_id = raw_id
    for i in cls_reacs:
        if i.get_enzyme_id() == tmp_id:
            return i


def reaction_titles(raw_id: str, cls_reacs: Optional[List[Reaction]] = None,
                    class_reaction: Optional[Reaction] = None,
                    pairs=None) -> Union[None, str]:
    fa_spec = False
    if "\nFA: " in raw_id:
        fa_spec = True
    if class_reaction is None:
        if cls_reacs is None:
            raise ValueError("Either 'class_reacs' or 'class_reaction' have to be specified!")
        class_reaction = get_class_reaction(raw_id, cls_reacs)
        if class_reaction is None:
            return None
    reaction_str = class_reaction.short_str(nice=True)
    if fa_spec:
        reaction_str += "\nFA: " + raw_id.split("\nFA: ")[1]
    if reaction_str == "":
        reaction_str = raw_id
    if pairs:
        reaction_str += "\nPair: " + pairs
    return reaction_str


def compute_ratios(unique_enzyme_list: List[str],
                   enzyme_structure_dict: Dict[str, str],
                   gr: nx.Graph,
                   reverse_mapping: Dict[str, str],
                   user_data: pd.DataFrame,
                   class_reactions: List[Reaction],
                   edge_attr: str,
                   difference_as_change: bool
                   ) -> Dict[str, pd.DataFrame]:
    reaction_ids: Dict[str, List[List[Union[str, Dict]]]] = {}
    multi_reaction_enzymes: Dict[str, List[str]] = {}
    for ed in gr.edges:
        tmp_edge_data = gr.get_edge_data(*ed)
        if tmp_edge_data[edge_attr] in unique_enzyme_list:
            if tmp_edge_data["reaction_id"] in reaction_ids.keys():
                reaction_ids[tmp_edge_data["reaction_id"]].append([ed[0], ed[1], tmp_edge_data])
            else:
                reaction_ids[tmp_edge_data["reaction_id"]] = [[ed[0], ed[1], tmp_edge_data]]
            if tmp_edge_data['reaction_structure'] != '1,1':
                # storing reaction id of multi-edge reactions
                multi_reaction_enzymes.setdefault(tmp_edge_data[edge_attr], []).append(tmp_edge_data['reaction_id'])

    df_list = {}
    for reac in unique_enzyme_list:
        counter = 0
        ratio_df = pd.DataFrame({})
        enzyme_struct = enzyme_structure_dict[reac].split(';')
        # length of 1 would be fatty acid modification => penalty
        # 1 substrate 1 product reaction
        if enzyme_struct[0] == '1,1':
            for edge in gr.edges:
                if gr.edges[edge][edge_attr] == reac and gr.edges[edge]['reaction_type'] != 'L_FAmodify':
                    tmp_edge = [reverse_mapping[edge[0]], reverse_mapping[edge[1]]]
                    if difference_as_change:
                        tmp_ratio = user_data[tmp_edge[0]] - user_data[tmp_edge[1]]
                    else:
                        tmp_ratio = user_data[tmp_edge[0]] / user_data[tmp_edge[1]]
                    if counter == 0:
                        ratio_df = pd.DataFrame({f'{tmp_edge[0]} - {tmp_edge[1]}': tmp_ratio})
                        counter += 1
                    else:
                        ratio_df = ratio_df.assign(**{f'{tmp_edge[0]} - {tmp_edge[1]}': tmp_ratio})
                        counter += 1
            df_list[reaction_titles(reac, class_reactions)] = ratio_df.copy()
        # multi substrate/product reaction
        else:
            for reaction in multi_reaction_enzymes[reac]:
                edge_list = reaction_ids[reaction]
                substrates = []
                products = []
                for edge in edge_list:
                    if edge[2]['l1_type'] == 'substrate':
                        substrates.append(user_data[reverse_mapping[edge[0]]])
                        products.append(user_data[reverse_mapping[edge[1]]])
                    else:
                        substrates.append(user_data[reverse_mapping[edge[1]]])
                        products.append(user_data[reverse_mapping[edge[0]]])
                if difference_as_change:
                    tmp_ratio = sum(products) - sum(substrates)
                else:
                    tmp_ratio = math.prod(products) / math.prod(substrates)
                if counter == 0:
                    ratio_df = pd.DataFrame({reaction: tmp_ratio})
                    counter += 1
                else:
                    ratio_df = ratio_df.assign(**{reaction: tmp_ratio})
                    counter += 1
            df_list[reaction_titles(reac, class_reactions)] = ratio_df.copy()
    return df_list


def plot_lineplot(data: Dict[str, pd.DataFrame],
                  plot_n_col: int = 3,
                  plot_fig_size: Tuple[int, int] = (10, 18),
                  plot_hspace: float = 0.4,) -> plt.Figure:
    def plot_line(ax, data, key):
        sns.lineplot(ax=ax,
                     data=data,
                     x="variable",
                     y="FC",
                     marker="o",
                     hue="groups"
                     ).set_title(key)

        ax.set(xlabel='Total chain length', ylabel='Abundance')

    return plot_template(
        data, plot_line, plot_n_col, plot_fig_size,
        plot_hspace
    )


def plot_networks(networks: Dict[str, nx.Graph],
                  scores: Dict[str, float],
                  p_values: Dict[str, float],
                  plot_n_col: int = 3,
                  plot_fig_size: Tuple[int, int] = (10, 18),
                  plot_hspace: float = 0.4) -> Tuple[Dict[str, nx.Graph], plt.Figure]:
    def plot_net(ax, data, key, scores, p_values):
        pos = nx.spring_layout(data)
        # labels = nx.get_edge_attributes(network, 'objective')
        # for key in labels:
        #     labels[key] = np.round(labels[key], 2)
        # nx.draw_networkx(data, pos, font_size=8, ax=ax)
        # nx.draw_networkx_edge_labels(network, pos, edge_labels=labels)
        lipid_nodes = [node for node in data.nodes
                       if data.nodes[node]['node_molecule_type'] == 'Lipid']
        nx.draw_networkx_nodes(
            data,
            pos,
            nodelist=lipid_nodes,
            node_shape='o',
            node_color=[lipid_colours[data.nodes[node]['lipid_class']]
                        for node in lipid_nodes],
            ax=ax
        )
        nx.draw_networkx_nodes(
            data,
            pos,
            nodelist=[node for node in data.nodes
                      if data.nodes[node]['node_molecule_type'] != 'Lipid'],
            node_shape='^',
            node_color='grey',
            ax=ax
        )
        nx.draw_networkx_edges(
            data, pos,
            edge_color='lightgrey',
            style=['solid' if not data.edges[edge].get('enzyme_edge', False) else 'dashed' for edge in data.edges],
            ax=ax
        )
        nx.draw_networkx_labels(
            data,
            pos,
            labels={
                node: data.nodes[node]['data_name'] if data.nodes[node]['node_molecule_type'] == 'Lipid'
                else data.nodes[node]['representation_id']
                for node in data.nodes
            },
            font_size=7, ax=ax
        )
        ax.set_title(key + f', score: {np.round(np.max(scores[key]), 3)}, p-value: {p_values[key]}')
        ax.axis('off')

    for name, net in networks.items():
        colours = {node: lipid_colours.get(net.nodes[node].get('lipid_class'), 'grey')
                   for node in net.nodes}
        shapes = {node: 'dot' if net.nodes[node]['node_molecule_type'] == 'lipid_class' else '^'
                  for node in net.nodes}
        labels = {node: node for node in net.nodes}
        nx.set_node_attributes(net, colours, "color")
        nx.set_node_attributes(net, shapes, "shape")
        nx.set_node_attributes(net, labels, "label")
        networks[name] = net

    return networks, plot_template(networks,
                                   plot_net,
                                   plot_n_col,
                                   plot_fig_size,
                                   plot_hspace,
                                   scores=scores,
                                   p_values=p_values)


def plot_template(data: Dict[str, any],
                  plot_function,
                  plot_n_col: int = 3,
                  plot_fig_size: Tuple[int, int] = (10, 18),
                  plot_hspace: float = 0.4,
                  **kwargs) -> plt.Figure:
    if len(data) == 1:
        fig, ax = plt.subplots(figsize=plot_fig_size)
        for key, ratio_df in data.items():
            plot_function(
                ax=ax,
                data=ratio_df,
                key=key,
                **kwargs
            )
        return fig

    columns = plot_n_col
    rows = math.ceil(len(data) / columns)
    single_row = rows < 2 or columns < 2
    rows = rows if rows > 0 else 1
    columns = columns if columns > 0 else 1
    fig, ax = plt.subplots(nrows=rows,
                           ncols=columns, figsize=plot_fig_size)
    fig.subplots_adjust(wspace=0.2,
                        hspace=plot_hspace)

    # loop over all dfs
    counter = 0
    for key, ratio_df in data.items():
        if single_row:
            plot_function(
                ax=ax[counter],
                data=ratio_df,
                key=key,
                **kwargs
            )
        else:
            plot_function(
                ax=ax[counter // columns, counter % columns],
                data=ratio_df,
                key=key,
                **kwargs
            )
        # At the end of the loop
        counter += 1
    # Delete last Axes if odd number of DataFrames
    while counter < ax.size:
        if single_row:
            fig.delaxes(ax[counter])
        else:
            fig.delaxes(ax[counter // columns, counter % columns])
        counter += 1

    return fig


def _add_missing_rows_(
    data: pd.DataFrame, species: List[str]
) -> pd.DataFrame:
    # test if this is really efficient or if we should
    # another way of adding rows
    dummy = pd.Series(index=data.columns)
    full_data = data.T.to_dict()
    for spec in species:
        if full_data.get(spec) is None:
            full_data[spec] = dummy
    return pd.DataFrame(full_data).T




from .reference import ReferenceLipid
from typing import List
import warnings

from . import reaction_types

__reaction_types__ = [x for x in dir(reaction_types) if not x.startswith("_")]


def reaction_validator(reaction_type: str, substrates: List[ReferenceLipid], products: List[ReferenceLipid],
                       verbose=True) -> bool:

    if (reaction_type is None) or (reaction_type == "") or (reaction_type not in __reaction_types__):
        raise KeyError("reaction type not (correctly) specified")

    # No lipid on one or both sides
    if (len(substrates) == 0) or (len(products) == 0):
        if verbose:
            warnings.warn("Boundary reaction. Lipids on at least one site missing.")
            print("Boundary reaction. Lipids on at least one site missing.")
        return True

    # One lipid on each side
    elif (len(substrates) == 1) and (len(products) == 1):
        if reaction_type == "L_HGmodify":
            return (substrates[0].get_fas() == products[0].get_fas()) \
                   and (substrates[0].get_ether() == products[0].get_ether()) \
                   and (substrates[0].get_lcb() == products[0].get_lcb()) \
                   and (substrates[0].get_potentialfas() == products[0].get_potentialfas()) \
                   and (substrates[0].get_headgroup() != products[0].get_headgroup())

        elif reaction_type == "L_HGdelete":
            return (substrates[0].get_fas() == products[0].get_fas()) \
                   and (substrates[0].get_ether() == products[0].get_ether()) \
                   and (substrates[0].get_lcb() == products[0].get_lcb()) \
                   and (((substrates[0].get_headgroup() != "") and (products[0].get_headgroup() == ""))
                        or ((substrates[0].get_headgroup() == "") and (products[0].get_headgroup() != "")))

        elif reaction_type == "L_FAdelete":
            return (substrates[0].get_ether() == products[0].get_ether()) \
                   and (substrates[0].get_lcb() == products[0].get_lcb()) \
                   and (substrates[0].get_headgroup() == products[0].get_headgroup()) \
                   and (((substrates[0].get_fas()+1) == products[0].get_fas())
                        or (substrates[0].get_fas() == (products[0].get_fas()+1)))
            # and (substrates[0].get_potentialfas() == products[0].get_potentialfas()) \
        elif reaction_type == "L_merge":
            if verbose:
                warnings.warn("Reaction with one product and one substrate defined as 'L_merge'")
                print("Reaction with one product and one substrate defined as 'L_merge'")
            return False

        elif reaction_type == "L_FAether":
            return (((substrates[0].get_ether()-1) == products[0].get_ether())
                    or (substrates[0].get_ether() == (products[0].get_ether()))-1) \
                   and (substrates[0].get_lcb() == products[0].get_lcb()) \
                   and (substrates[0].get_potentialfas() == products[0].get_potentialfas()) \
                   and (substrates[0].get_headgroup() == products[0].get_headgroup()) \
                   and ((substrates[0].get_fas()) == products[0].get_fas())
        else:
            if verbose:
                warnings.warn("Reaction " + reaction_type +
                              " not implemented for a reaction with one substrate and one product.")
                print("Reaction " + reaction_type +
                      " not implemented for a reaction with one substrate and one product.")
            return False

    # 2&1 or 1&2
    elif ((len(substrates) == 1) and (len(products) == 2)) or ((len(substrates) == 2) and (len(products) == 1)):
        twos = substrates if len(substrates) == 2 else products
        ones = substrates if len(substrates) == 1 else products

        if reaction_type == "L_FAdelete":
            return (ones[0].get_fas() == sum([x.get_fas() for x in twos])) \
                   and any([x.get_headgroup() == ones[0].get_headgroup() for x in twos]) \
                   and any([x.get_potentialfas() > x.get_fas() for x in twos])

        elif reaction_type == "L_HGdelete":
            return (ones[0].get_fas() == sum([x.get_fas() for x in twos])) \
                   and any([x.get_headgroup() == ones[0].get_headgroup() for x in twos])

        elif reaction_type == "L_merge":
            return ones[0].get_fas() == sum([x.get_fas() for x in twos])  # TODO: Check LCB and Ether

        else:
            if verbose:
                warnings.warn(reaction_type + " not implemented for (1,2) or (2,1) reactions")
                print(reaction_type + " not implemented for (1,2) or (2,1) reactions")
            return False

    # 2&2
    elif (len(substrates) == 2) and (len(products) == 2):
        if reaction_type == "L_FAdelete":
            return (sum([x.get_fas() for x in substrates]) == sum([x.get_fas() for x in products])) \
                   and (all([x.get_headgroup() in [y.get_headgroup() for y in products] for x in substrates]))

        elif reaction_type in ["L_HGdelete", "L_HGmodify"]:
            return (sum([x.get_fas() for x in substrates]) == sum([x.get_fas() for x in products])) \
                   and (all([x.get_fas() in [y.get_fas() for y in products] for x in substrates]))

        else:
            if verbose:
                warnings.warn(reaction_type + " not implemented for (2,2) reactions")
                print(reaction_type + " not implemented for (2,2) reactions")
            return False

    else:
        raise NotImplementedError(reaction_type + " with " + str([len(substrates), len(products)]) + ", but the only possible combinations of substrate product pairs are currently "
                                  "(0,1);(1,0);(1,1);(1,2);(2,1);(2,2)")
